from __future__ import annotations

import enum
import logging

from prettyqt import constants, core, itemmodels
from prettyqt.utils import classhelpers


logger = logging.getLogger(__name__)


class FakeClassNameProp:
    def isWritable(self):
        return False

    def get_name(self):
        return "Class"

    def read(self, qobject):
        return qobject.__class__.__name__

    def write(self, qobject, value):
        pass


class FakeUserPropertyNameProp:
    def isWritable(self):
        return False

    def get_name(self):
        return "User property name"

    def read(self, qobject):
        userprop = qobject.metaObject().userProperty()
        return userprop.name() if userprop.isValid() else ""

    def write(self, qobject, value):
        pass


class FakeUserPropertyValueProp:
    def isWritable(self):
        return True

    def get_name(self):
        return "User property value"

    def read(self, qobject):
        userprop = qobject.metaObject().userProperty()
        return userprop.read(qobject)

    def write(self, qobject, value):
        userprop = qobject.metaObject().userProperty()
        return userprop.write(qobject, value)


class FakeLayoutProp:
    def isWritable(self):
        return False

    def get_name(self):
        return "Layout"

    def read(self, qobject):
        # check for layout attr to support QObjects
        if not hasattr(qobject, "layout") or qobject.layout() is None:
            return ""
        return qobject.layout().__class__.__name__

    def write(self, qobject, value):
        pass


class BaseHierarchyModel(itemmodels.TreeModel):
    class Roles(enum.IntEnum):
        """Custom roles."""

        WidgetRole = constants.USER_ROLE + 23324

    def __init__(
        self,
        obj: core.QObject,
        show_root: bool = True,
        parent: core.QObject | None = None,
    ):
        self.BaseClass = None
        self.props = None
        super().__init__(obj=obj, show_root=show_root, parent=parent)
        self.set_base_class("QWidget")

    def set_base_class(self, klass: str | type[core.QObject]):
        from prettyqt import widgets

        if isinstance(klass, str):
            for kls in classhelpers.get_subclasses(core.QObject):
                if kls.__name__ == klass:
                    klass = kls
                    break
            else:
                klass = core.QObject
        with self.reset_model():
            self.BaseClass = klass
            metaobj = klass.staticMetaObject
            self.props = core.MetaObject(metaobj).get_properties(only_stored=True)
            self.props.insert(0, FakeClassNameProp())
            if issubclass(klass, widgets.QWidget):
                self.props.insert(1, FakeLayoutProp())
            self.props.insert(1, FakeUserPropertyNameProp())
            self.props.insert(1, FakeUserPropertyValueProp())

    def get_base_class_name(self) -> str:
        return self.BaseClass.__name__

    def get_base_class(self) -> type[core.QObject]:
        return self.BaseClass

    def columnCount(self, parent=None):
        return len(self.props)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.props[section].get_name()

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        widget = self.data_by_index(index).obj
        prop = self.props[index.column()]
        val = prop.read(widget)
        is_bool = isinstance(val, bool)
        match role:
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE if not is_bool:
                return val
            case constants.CHECKSTATE_ROLE if is_bool:
                return self.to_checkstate(val)
            case constants.USER_ROLE:
                prop = self.props[index.column()]
                return val
            case self.Roles.WidgetRole:
                return widget
            case constants.SIZE_HINT_ROLE:
                return core.QSize(250, 35)

    def setData(
        self,
        index: core.ModelIndex,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        prop = self.props[index.column()]
        widget = self.data_by_index(index).obj
        match role:
            case constants.EDIT_ROLE | constants.USER_ROLE:
                with self.change_layout():
                    prop.write(widget, value)
                # self.update_row(index.row())
                return True
            case constants.CHECKSTATE_ROLE:
                with self.change_layout():
                    prop.write(widget, bool(value))
                return True
        return False

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        prop = self.props[index.column()]
        widget = self.data_by_index(index).obj
        val = prop.read(widget)
        flag = constants.IS_CHECKABLE if isinstance(val, bool) else constants.IS_EDITABLE
        default = super().flags(index)
        return (default | flag) if prop.isWritable() else default

    base_class_name = core.Property(
        str,
        get_base_class_name,
        set_base_class,
        doc="Class to use for the columns.",
    )
    # base_class = core.Property(type, get_base_class, set_base_class, stored=False)


class WidgetHierarchyModel(BaseHierarchyModel):
    """Tree model showing the child tree of a QObject."""

    SUPPORTS = core.QObject

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, core.QObject)

    def _fetch_object_children(
        self, item: WidgetHierarchyModel.TreeItem
    ) -> list[WidgetHierarchyModel.TreeItem]:
        return [
            self.TreeItem(obj=i)
            for i in item.obj.findChildren(
                self.BaseClass, None, constants.FindChildOption.FindDirectChildrenOnly
            )
        ]

    def _has_children(self, item: WidgetHierarchyModel.TreeItem) -> bool:
        return bool(
            item.obj.findChildren(
                self.BaseClass, None, constants.FindChildOption.FindDirectChildrenOnly
            )
        )


class LayoutHierarchyModel(BaseHierarchyModel):
    """Tree model showing the layout hierarchy of widgets."""

    SUPPORTS = "widgets.QWidget"

    @classmethod
    def supports(cls, instance) -> bool:
        from prettyqt import widgets

        return isinstance(instance, widgets.QWidget)

    def _fetch_object_children(
        self, item: LayoutHierarchyModel.TreeItem
    ) -> list[LayoutHierarchyModel.TreeItem]:
        from prettyqt import widgets

        match item.obj:
            case (
                widgets.QSplitter()
                | widgets.QToolBox()
                | widgets.QStackedWidget()
                | widgets.QTabWidget()
            ):
                items = [item.obj.widget(i) for i in range(item.obj.count())]
            case widgets.QWidget():
                layout = item.obj.layout()
                items = [layout.itemAt(i) for i in range(layout.count())]
                items = [w if (w := i.widget()) else i.layout() for i in items]
            case widgets.QLayout():
                layout = item.obj
                items = [layout.itemAt(i) for i in range(layout.count())]
                items = [w if (w := i.widget()) else i.layout() for i in items]
            case _:
                raise ValueError(item)
        return [self.TreeItem(obj=i) for i in items]

    def _has_children(self, item: LayoutHierarchyModel.TreeItem) -> bool:
        from prettyqt import widgets

        match item.obj:
            case (
                widgets.QSplitter()
                | widgets.QToolBox()
                | widgets.QStackedWidget()
                | widgets.QTabWidget()
                | widgets.QLayout()
            ):
                return item.obj.count() > 0
            case _:
                layout = item.obj.layout()
                return False if layout is None else layout.count() > 0


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    view = widgets.TreeView()
    view.setRootIsDecorated(True)
    widget = debugging.example_widget()

    model = WidgetHierarchyModel(widget, parent=view)
    view = debugging.ProxyComparerWidget(model, itemview="tree")
    # view.set_model(model)
    # view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    # view.set_delegate("editor")
    # view.resize(1000, 1000)
    view.show()
    # widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import math
import os

from prettyqt import constants, core, gui, widgets


def fit_image(
    width: float, height: float, pwidth: float, pheight: float
) -> tuple[bool, int, int]:
    """Fit image in box of width pwidth and height pheight.

    Arguments:
        width: Width of image
        height: Height of image
        pwidth: Width of box
        pheight: Height of box

    Returns:
        scaled, new_width, new_height.
    """
    if height < 1 or width < 1:
        return False, int(width), int(height)
    scaled = height > pheight or width > pwidth
    if height > pheight:
        corrf = pheight / height
        width, height = math.floor(corrf * width), pheight
    if width > pwidth:
        corrf = pwidth / float(width)
        width, height = pwidth, math.floor(corrf * height)
    if height > pheight:
        corrf = pheight / float(height)
        width, height = math.floor(corrf * width), pheight
    return scaled, int(width), int(height)


def draw_size(p: gui.Painter, rect: core.Rect, w: int, h: int):
    rect = rect.adjusted(0, 0, 0, -4)
    f = p.font()
    f.setBold(True)
    p.setFont(f)
    sz = "\u00a0%d x %d\u00a0" % (w, h)
    flags = (
        constants.AlignmentFlag.AlignBottom
        | constants.AlignmentFlag.AlignRight
        | constants.TextFlag.TextSingleLine
    )
    szrect = p.boundingRect(rect, flags, sz)
    p.fillRect(szrect.adjusted(0, 0, 0, 4), gui.Color(0, 0, 0, 200))
    p.setPen(gui.Pen(gui.Color(255, 255, 255)))
    p.drawText(rect, flags, sz)


class ImageViewer(widgets.Widget):
    def __init__(
        self,
        image: gui.QPixmap | os.PathLike | None = None,
        *,
        show_border: bool = True,
        show_size: bool = False,
        border_width: int = 1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._border_width = border_width
        self._pixmap = gui.Pixmap()
        self.set_image(image)
        self.setMinimumSize(core.Size(150, 200))
        self.draw_border = show_border
        self.show_size = show_size

    def set_image(self, pixmap: gui.QPixmap | os.PathLike | None):
        match pixmap:
            case gui.QPixmap():
                self._pixmap = pixmap
            case os.PathLike():
                self._pixmap = gui.Pixmap.from_file(pixmap)
            case None:
                self._pixmap = gui.Pixmap()
        self.updateGeometry()
        self.update()

    def toggle_show_size(self):
        self.show_size ^= True
        self.update()

    def pixmap(self):
        return self._pixmap

    def sizeHint(self) -> core.QSize:
        return self.minimumSize() if self._pixmap.isNull() else self._pixmap.size()

    def paintEvent(self, event):
        super().paintEvent(event)
        pmap = self._pixmap
        w, h = pmap.width(), pmap.height()
        ow, oh = w, h
        cw, ch = self.rect().width(), self.rect().height()
        scaled, nw, nh = fit_image(w, h, cw, ch)
        if scaled:
            pmap = pmap.scaled(
                int(nw * pmap.devicePixelRatio()),
                int(nh * pmap.devicePixelRatio()),
                constants.AspectRatioMode.IgnoreAspectRatio,
                constants.TransformationMode.SmoothTransformation,
            )
        w = pmap.width() // pmap.devicePixelRatio()
        h = pmap.height() // pmap.devicePixelRatio()

        x = abs(cw - w) // 2
        y = abs(ch - h) // 2
        target = core.Rect(int(x), int(y), int(w), int(h))
        with gui.Painter(self) as p:
            p.setRenderHints(
                gui.Painter.RenderHint.Antialiasing
                | gui.Painter.RenderHint.SmoothPixmapTransform
            )
            p.drawPixmap(target, pmap)
            if self.draw_border:
                pen = gui.Pen()
                pen.setWidth(self._border_width)
                p.setPen(pen)
                p.drawRect(target)
            if self.show_size:
                draw_size(p, target, ow, oh)


if __name__ == "__main__":
    from prettyqt import iconprovider

    app = widgets.app()
    widget = ImageViewer()
    icon = iconprovider.get_icon("mdi.folder")
    widget.set_image(icon.pixmap(256))
    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict


GestureCancelPolicyStr = Literal["none", "all_in_context"]

GESTURE_CANCEL_POLICY: bidict[
    GestureCancelPolicyStr, widgets.QGesture.GestureCancelPolicy
] = bidict(
    none=widgets.QGesture.GestureCancelPolicy.CancelNone,
    all_in_context=widgets.QGesture.GestureCancelPolicy.CancelAllInContext,
)


class GestureMixin(core.ObjectMixin):
    def get_state(self) -> constants.GestureStateStr:
        """Return current state.

        Returns:
            state
        """
        return constants.GESTURE_STATE.inverse[self.state()]

    def get_gesture_type(self) -> constants.GestureTypeStr:
        """Return current gesture type.

        Returns:
            gesture type
        """
        return constants.GESTURE_TYPE.inverse[self.gestureType()]

    def get_hot_spot(self) -> core.PointF:
        return core.PointF(self.hotSpot())

    def set_gesture_cancel_policy(
        self, policy: GestureCancelPolicyStr | widgets.QGesture.GestureCancelPolicy
    ):
        """Set gesture cancel policy.

        Args:
            policy: gesture cancel policy to use
        """
        self.setGestureCancelPolicy(GESTURE_CANCEL_POLICY.get_enum_value(policy))

    def get_gesture_cancel_policy(self) -> GestureCancelPolicyStr:
        """Return current gesture cancel policy.

        Returns:
            gesture cancel policy
        """
        return GESTURE_CANCEL_POLICY.inverse[self.gestureCancelPolicy()]


class Gesture(GestureMixin, widgets.QGesture):
    """A gesture, containing properties that describe the corresponding user input."""


if __name__ == "__main__":
    app = widgets.app()
    gesture = Gesture()



from __future__ import annotations

from collections.abc import Callable, Iterable
import dataclasses

from prettyqt import constants


@dataclasses.dataclass
class Command:
    keycombo: str
    description: str
    context: str
    method: Callable


_registry = dict()


def register(
    keybinding: str | Iterable[str], command: str
) -> Callable[[Callable], Callable]:
    def decorator(fn: Callable) -> Callable:
        if isinstance(keybinding, str):
            bind(keybinding, command, fn)
        else:
            for binding in keybinding:
                bind(binding, command, fn, allow_override=False)
        return fn

    return decorator


def bind(
    keybinding: str,
    command: str,
    fn: Callable,
    mode: constants.ShortcutContextStr = "application",
    allow_override: bool = True,
) -> None:
    if not allow_override and keybinding in _registry:
        raise ValueError(f"Duplicate keybinding for '{keybinding}'")
    _registry[keybinding] = (command, fn)


def unbind(keybinding: str, mode: constants.ShortcutContextStr = "application") -> None:
    """Remove keybinding from registry.

    See config/configcommands.unbind for the corresponding command.
    """
    if keybinding not in _registry:
        raise KeyError(f"No binding found for '{keybinding}'")
    del _registry[keybinding]



from __future__ import annotations

import contextlib

from typing import Literal

from prettyqt import gui
from prettyqt.qt import QtCore, QtQuick
from prettyqt.utils import bidict


CreateTextureOptionStr = Literal[
    "has_alpha_channel",
    "has_mipmaps",
    "owns_gl_texture",
    "can_use_atlas",
    "is_opaque",
]

CREATE_TEXTURE_OPTION: bidict[
    CreateTextureOptionStr, QtQuick.QQuickWindow.CreateTextureOption
] = bidict(
    has_alpha_channel=QtQuick.QQuickWindow.CreateTextureOption.TextureHasAlphaChannel,
    has_mipmaps=QtQuick.QQuickWindow.CreateTextureOption.TextureHasMipmaps,
    owns_gl_texture=QtQuick.QQuickWindow.CreateTextureOption.TextureOwnsGLTexture,
    can_use_atlas=QtQuick.QQuickWindow.CreateTextureOption.TextureCanUseAtlas,
    is_opaque=QtQuick.QQuickWindow.CreateTextureOption.TextureIsOpaque,
)

RenderStageStr = Literal[
    "before_synchronizing",
    "after_synchronizing",
    "before_rendering",
    "after_rendering",
    "after_swap",
    "no_stage",
]

RENDER_STAGE: bidict[RenderStageStr, QtQuick.QQuickWindow.RenderStage] = bidict(
    before_synchronizing=QtQuick.QQuickWindow.RenderStage.BeforeSynchronizingStage,
    after_synchronizing=QtQuick.QQuickWindow.RenderStage.AfterSynchronizingStage,
    before_rendering=QtQuick.QQuickWindow.RenderStage.BeforeRenderingStage,
    after_rendering=QtQuick.QQuickWindow.RenderStage.AfterRenderingStage,
    after_swap=QtQuick.QQuickWindow.RenderStage.AfterSwapStage,
    no_stage=QtQuick.QQuickWindow.RenderStage.NoStage,
)

TextRenderTypeStr = Literal["qt_text", "native_text"]

TEXT_RENDER_TYPE: bidict[TextRenderTypeStr, QtQuick.QQuickWindow.TextRenderType] = bidict(
    qt_text=QtQuick.QQuickWindow.TextRenderType.QtTextRendering,
    native_text=QtQuick.QQuickWindow.TextRenderType.NativeTextRendering,
)


class QuickWindowMixin(gui.WindowMixin):
    def create_texture_from_image(self, image: gui.QImage, **kwargs):
        flag = self.CreateTextureOption(0)
        for key, val in kwargs.items():
            if val is True:
                v = CREATE_TEXTURE_OPTION[key]
                flag |= v
        return self.createTextureFromImage(image, flag)  # type: ignore

    def grab_window(self) -> gui.Image:
        return gui.Image(self.grabWindow())

    def get_color(self) -> gui.Color:
        return gui.Color(self.color())

    @classmethod
    def set_text_render_type(cls, typ: TextRenderTypeStr):
        """Set the default render type of text-like elements in Qt Quick.

        Note: setting the render type will only affect elements created afterwards;
        the render type of existing elements will not be modified.

        Args:
            typ: text render type to use
        """
        cls.setTextRenderType(TEXT_RENDER_TYPE.get_enum_value(typ))

    @classmethod
    def get_text_render_type(cls) -> TextRenderTypeStr:
        """Return the render type of text-like elements in Qt Quick.

        Returns:
            text render type
        """
        return TEXT_RENDER_TYPE.inverse[cls.textRenderType()]

    @contextlib.contextmanager
    def external_commands(self):
        self.beginExternalCommands()
        yield self
        self.endExternalCommands()

    def schedule_render_job(
        self,
        job: QtCore.QRunnable,
        render_stage: RenderStageStr | QtQuick.QQuickWindow.RenderStage,
    ):
        self.scheduleRenderJob(job, RENDER_STAGE.get_enum_value(render_stage))


class QuickWindow(QuickWindowMixin, QtQuick.QQuickWindow):
    """The window for displaying a graphical QML scene."""


if __name__ == "__main__":
    app = gui.app()
    wnd = QuickWindow()
    wnd.set_text_render_type("qt_text")
    img = gui.QImage()
    texture = wnd.create_texture_from_image(img)



# credits to https://stackoverflow.com/a/54819051

from __future__ import annotations

from collections.abc import Sequence

from prettyqt import constants, core, gui, widgets


class LabeledSlider(widgets.Widget):
    value_changed = core.Signal(str)

    def __init__(
        self,
        labels: Sequence[str],
        orientation: constants.OrientationStr | constants.Orientation = "horizontal",
        parent: widgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)

        if not isinstance(labels, Sequence):
            raise ValueError("<labels> must be a sequence.")
        self.levels = labels
        orientation = constants.ORIENTATION.get_enum_value(orientation)
        self.set_layout(constants.ORIENTATION.inverse[orientation], margin=10)

        # gives some space to print labels
        self.left_margin = 10
        self.top_margin = 10
        self.right_margin = 10
        self.bottom_margin = 10

        self.sl = widgets.Slider(
            orientation,
            tick_interval=1,
            single_step=1,
            minimum=0,
            maximum=len(self.levels) - 1,
            value=0,
        )
        self.sl.value_changed.connect(self._on_value_change)
        if orientation == constants.HORIZONTAL:
            self.sl.set_tick_position("below")
            self.sl.setMinimumWidth(300)
        else:
            self.sl.set_tick_position("left")
            self.sl.setMinimumHeight(300)

        self.box.add(self.sl)

    def _on_value_change(self, value: int):
        self.value_changed.emit(self.levels[value])

    def paintEvent(self, e):
        super().paintEvent(e)

        style = self.sl.style()
        st_slider = widgets.StyleOptionSlider()
        st_slider.initFrom(self.sl)
        st_slider.orientation = self.sl.orientation()

        length = style.pixelMetric(
            widgets.Style.PixelMetric.PM_SliderLength, st_slider, self.sl
        )
        available = style.pixelMetric(
            widgets.Style.PixelMetric.PM_SliderSpaceAvailable, st_slider, self.sl
        )

        with gui.Painter(self) as painter:
            for v, v_str in enumerate(self.levels):
                # get the size of the label
                rect = painter.get_text_rect(v_str)

                if self.sl.is_horizontal():
                    x_loc = widgets.Style.sliderPositionFromValue(
                        self.sl.minimum(), self.sl.maximum(), v, available
                    )
                    # I assume the offset is half the length of slider, therefore
                    # + length//2
                    x_loc += length // 2

                    # left bound of the text = center - half of text width + L_margin
                    left = x_loc - rect.width() // 2 + self.left_margin
                    bottom = self.rect().bottom()
                    # enlarge margins if clipping
                    if v == self.sl.minimum():
                        if left <= 0:
                            self.left_margin = rect.width() // 2 - x_loc
                        self.bottom_margin = max(self.bottom_margin, rect.height())
                        self.adjust_margins()
                    if v == self.sl.maximum() and rect.width() // 2 >= self.right_margin:
                        self.right_margin = rect.width() // 2
                        self.adjust_margins()
                else:
                    y_loc = widgets.Style.sliderPositionFromValue(
                        self.sl.minimum(),
                        self.sl.maximum(),
                        v,
                        available,
                        upsideDown=True,
                    )

                    bottom = (
                        y_loc + length // 2 + rect.height() // 2 + self.top_margin - 3
                    )
                    # there is a 3 px offset that I can't attribute to any metric

                    left = self.left_margin - rect.width()
                    if left <= 0:
                        self.left_margin = rect.width() + 2
                        self.adjust_margins()
                painter.drawText(left, bottom, v_str)

    @classmethod
    def setup_example(cls):
        return cls(labels=["a", "b", "c"])

    def adjust_margins(self):
        self.box.setContentsMargins(
            self.left_margin,
            self.top_margin,
            self.right_margin,
            self.bottom_margin,
        )


if __name__ == "__main__":
    app = widgets.app()
    frame = widgets.Widget()
    ha = widgets.HBoxLayout()
    frame.setLayout(ha)

    w = LabeledSlider(labels=["test", "test2", "test3"], orientation="vertical")
    w.value_changed.connect(print)
    ha.addWidget(w)
    frame.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterable

from prettyqt import constants, core, gui


class Drag(core.ObjectMixin, gui.QDrag):
    """Support for MIME-based drag and drop data transfer."""

    def get_pixmap(self) -> gui.Pixmap:
        return gui.Pixmap(self.pixmap())

    def get_default_action(self) -> constants.DropActionStr:
        return constants.DROP_ACTION.inverse[self.defaultAction()]

    def get_drag_cursor(
        self, action: constants.DropActionStr | constants.DropAction
    ) -> gui.Pixmap:
        px = self.dragCursor(constants.DROP_ACTION.get_enum_value(action))
        return gui.Pixmap(px)

    def set_drag_cursor(
        self,
        cursor: gui.QPixmap,
        action: constants.DropActionStr | constants.DropAction,
    ):
        self.setDragCursor(cursor, constants.DROP_ACTION.get_enum_value(action))

    def get_supported_actions(self) -> list[constants.DropActionStr]:
        return constants.DROP_ACTION.get_list(self.supportedActions())

    def main_loop(
        self,
        supported_actions: Iterable[constants.DropActionStr | constants.DropAction]
        | None = None,
        default_drop_action: constants.DropActionStr | constants.DropAction | None = None,
    ) -> constants.DropActionStr:
        supported_actions = supported_actions or [constants.DROP_ACTION["move"]]
        flag = constants.DropAction(0)
        for i in supported_actions:
            flag |= constants.DROP_ACTION.get_enum_value(i)
        default_action = constants.DROP_ACTION.get_enum_value(default_drop_action)
        result = super().exec(flag, default_action)
        return constants.DROP_ACTION.inverse[result]


if __name__ == "__main__":
    obj = core.Object()
    drag = Drag(obj)
    drag.get_pixmap()



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtWebEngineCore


class WebEngineUrlSchemeHandler(
    core.ObjectMixin, QtWebEngineCore.QWebEngineUrlSchemeHandler
):
    pass


if __name__ == "__main__":
    item = WebEngineUrlSchemeHandler()



from __future__ import annotations

from typing import Literal

from prettyqt.qt import QtGui
from prettyqt.utils import bidict, get_repr


PAGE_SIZE_ID = bidict(
    a0=QtGui.QPageSize.PageSizeId.A0,
    a1=QtGui.QPageSize.PageSizeId.A1,
    a2=QtGui.QPageSize.PageSizeId.A2,
    a3=QtGui.QPageSize.PageSizeId.A3,
    a4=QtGui.QPageSize.PageSizeId.A4,
    a5=QtGui.QPageSize.PageSizeId.A5,
    a6=QtGui.QPageSize.PageSizeId.A6,
    a7=QtGui.QPageSize.PageSizeId.A7,
    a8=QtGui.QPageSize.PageSizeId.A8,
    a9=QtGui.QPageSize.PageSizeId.A9,
    b0=QtGui.QPageSize.PageSizeId.B0,
    b1=QtGui.QPageSize.PageSizeId.B1,
    b2=QtGui.QPageSize.PageSizeId.B2,
    b3=QtGui.QPageSize.PageSizeId.B3,
    b4=QtGui.QPageSize.PageSizeId.B4,
    b5=QtGui.QPageSize.PageSizeId.B5,
    b6=QtGui.QPageSize.PageSizeId.B6,
    b7=QtGui.QPageSize.PageSizeId.B7,
    b8=QtGui.QPageSize.PageSizeId.B8,
    b9=QtGui.QPageSize.PageSizeId.B9,
    b10=QtGui.QPageSize.PageSizeId.B10,
    c5e=QtGui.QPageSize.PageSizeId.C5E,
    comm_10e=QtGui.QPageSize.PageSizeId.Comm10E,
    dle=QtGui.QPageSize.PageSizeId.DLE,
    executive=QtGui.QPageSize.PageSizeId.Executive,
    folio=QtGui.QPageSize.PageSizeId.Folio,
    ledger=QtGui.QPageSize.PageSizeId.Ledger,
    legal=QtGui.QPageSize.PageSizeId.Legal,
    letter=QtGui.QPageSize.PageSizeId.Letter,
    tabloid=QtGui.QPageSize.PageSizeId.Tabloid,
    custom=QtGui.QPageSize.PageSizeId.Custom,
    a10=QtGui.QPageSize.PageSizeId.A10,
    a3_extra=QtGui.QPageSize.PageSizeId.A3Extra,
    a4_extra=QtGui.QPageSize.PageSizeId.A4Extra,
    a4_plus=QtGui.QPageSize.PageSizeId.A4Plus,
    a4_small=QtGui.QPageSize.PageSizeId.A4Small,
    a5_extra=QtGui.QPageSize.PageSizeId.A5Extra,
    b5_extra=QtGui.QPageSize.PageSizeId.B5Extra,
    jis_b0=QtGui.QPageSize.PageSizeId.JisB0,
    jis_b1=QtGui.QPageSize.PageSizeId.JisB1,
    jis_b2=QtGui.QPageSize.PageSizeId.JisB2,
    jis_b3=QtGui.QPageSize.PageSizeId.JisB3,
    jis_b4=QtGui.QPageSize.PageSizeId.JisB4,
    jis_b5=QtGui.QPageSize.PageSizeId.JisB5,
    jis_b6=QtGui.QPageSize.PageSizeId.JisB6,
    jis_b7=QtGui.QPageSize.PageSizeId.JisB7,
    jis_b8=QtGui.QPageSize.PageSizeId.JisB8,
    jis_b9=QtGui.QPageSize.PageSizeId.JisB9,
    jis_b10=QtGui.QPageSize.PageSizeId.JisB10,
    # c5e=QtGui.QPageSize.PageSizeId.AnsiA, Letter
    # c5e=QtGui.QPageSize.PageSizeId.AnsiB, Ledger
    ansi_c=QtGui.QPageSize.PageSizeId.AnsiC,
    ansi_d=QtGui.QPageSize.PageSizeId.AnsiD,
    ansi_e=QtGui.QPageSize.PageSizeId.AnsiE,
    legal_extra=QtGui.QPageSize.PageSizeId.LegalExtra,
    letter_extra=QtGui.QPageSize.PageSizeId.LetterExtra,
    letter_plus=QtGui.QPageSize.PageSizeId.LetterPlus,
    letter_small=QtGui.QPageSize.PageSizeId.LetterSmall,
    tabloid_extra=QtGui.QPageSize.PageSizeId.TabloidExtra,
    arch_a=QtGui.QPageSize.PageSizeId.ArchA,
    arch_b=QtGui.QPageSize.PageSizeId.ArchB,
    arch_c=QtGui.QPageSize.PageSizeId.ArchC,
    arch_d=QtGui.QPageSize.PageSizeId.ArchD,
    arch_e=QtGui.QPageSize.PageSizeId.ArchE,
    imperial_7x9=QtGui.QPageSize.PageSizeId.Imperial7x9,
    imperial_8x10=QtGui.QPageSize.PageSizeId.Imperial8x10,
    imperial_9x11=QtGui.QPageSize.PageSizeId.Imperial9x11,
    imperial_9x12=QtGui.QPageSize.PageSizeId.Imperial9x12,
    imperial_10x11=QtGui.QPageSize.PageSizeId.Imperial10x11,
    imperial_10x13=QtGui.QPageSize.PageSizeId.Imperial10x13,
    imperial_10x14=QtGui.QPageSize.PageSizeId.Imperial10x14,
    imperial_12x11=QtGui.QPageSize.PageSizeId.Imperial12x11,
    imperial_15x11=QtGui.QPageSize.PageSizeId.Imperial15x11,
    executive_standard=QtGui.QPageSize.PageSizeId.ExecutiveStandard,
    note=QtGui.QPageSize.PageSizeId.Note,
    quarto=QtGui.QPageSize.PageSizeId.Quarto,
    statement=QtGui.QPageSize.PageSizeId.Statement,
    super_a=QtGui.QPageSize.PageSizeId.SuperA,
    super_b=QtGui.QPageSize.PageSizeId.SuperB,
    postcard=QtGui.QPageSize.PageSizeId.Postcard,
    double_postcard=QtGui.QPageSize.PageSizeId.DoublePostcard,
    prc_16_k=QtGui.QPageSize.PageSizeId.Prc16K,
    prc_32_k=QtGui.QPageSize.PageSizeId.Prc32K,
    prc_32_k_big=QtGui.QPageSize.PageSizeId.Prc32KBig,
    fan_fold_us=QtGui.QPageSize.PageSizeId.FanFoldUS,
    fan_fold_german=QtGui.QPageSize.PageSizeId.FanFoldGerman,
    fan_fold_german_legal=QtGui.QPageSize.PageSizeId.FanFoldGermanLegal,
    envelope_b4=QtGui.QPageSize.PageSizeId.EnvelopeB4,
    envelope_b5=QtGui.QPageSize.PageSizeId.EnvelopeB5,
    envelope_b6=QtGui.QPageSize.PageSizeId.EnvelopeB6,
    envelope_c0=QtGui.QPageSize.PageSizeId.EnvelopeC0,
    envelope_c1=QtGui.QPageSize.PageSizeId.EnvelopeC1,
    envelope_c2=QtGui.QPageSize.PageSizeId.EnvelopeC2,
    envelope_c3=QtGui.QPageSize.PageSizeId.EnvelopeC3,
    envelope_c4=QtGui.QPageSize.PageSizeId.EnvelopeC4,
    # envelope_c5=QtGui.QPageSize.PageSizeId.EnvelopeC5, c5e
    envelope_c6=QtGui.QPageSize.PageSizeId.EnvelopeC6,
    envelope_c65=QtGui.QPageSize.PageSizeId.EnvelopeC65,
    envelope_c7=QtGui.QPageSize.PageSizeId.EnvelopeC7,
    # envelope_dl=QtGui.QPageSize.PageSizeId.EnvelopeDL, dle
    envelope_9=QtGui.QPageSize.PageSizeId.Envelope9,
    # envelope_10=QtGui.QPageSize.PageSizeId.Envelope10, comm10e
    envelope_11=QtGui.QPageSize.PageSizeId.Envelope11,
    envelope_12=QtGui.QPageSize.PageSizeId.Envelope12,
    envelope_14=QtGui.QPageSize.PageSizeId.Envelope14,
    envelope_monarch=QtGui.QPageSize.PageSizeId.EnvelopeMonarch,
    envelope_personal=QtGui.QPageSize.PageSizeId.EnvelopePersonal,
    envelope_chou_3=QtGui.QPageSize.PageSizeId.EnvelopeChou3,
    envelope_chou_4=QtGui.QPageSize.PageSizeId.EnvelopeChou4,
    envelope_invite=QtGui.QPageSize.PageSizeId.EnvelopeInvite,
    envelope_italian=QtGui.QPageSize.PageSizeId.EnvelopeItalian,
    envelope_kaku_2=QtGui.QPageSize.PageSizeId.EnvelopeKaku2,
    envelope_kaku_3=QtGui.QPageSize.PageSizeId.EnvelopeKaku3,
    envelope_prc_1=QtGui.QPageSize.PageSizeId.EnvelopePrc1,
    envelope_prc_2=QtGui.QPageSize.PageSizeId.EnvelopePrc2,
    envelope_prc_3=QtGui.QPageSize.PageSizeId.EnvelopePrc3,
    envelope_prc_4=QtGui.QPageSize.PageSizeId.EnvelopePrc4,
    envelope_prc_5=QtGui.QPageSize.PageSizeId.EnvelopePrc5,
    envelope_prc_6=QtGui.QPageSize.PageSizeId.EnvelopePrc6,
    envelope_prc_7=QtGui.QPageSize.PageSizeId.EnvelopePrc7,
    envelope_prc_8=QtGui.QPageSize.PageSizeId.EnvelopePrc8,
    envelope_prc_9=QtGui.QPageSize.PageSizeId.EnvelopePrc9,
    envelope_prc_10=QtGui.QPageSize.PageSizeId.EnvelopePrc10,
    envelope_you_4=QtGui.QPageSize.PageSizeId.EnvelopeYou4,
    # last_page_size=QtGui.QPageSize.PageSizeId.LastPageSize, envelope_you_4
)

SIZE_MATCH_POLICY = bidict(
    fuzzy=QtGui.QPageSize.SizeMatchPolicy.FuzzyMatch,
    fuzzy_orientation=QtGui.QPageSize.SizeMatchPolicy.FuzzyOrientationMatch,
    exact=QtGui.QPageSize.SizeMatchPolicy.ExactMatch,
)

UNITS = bidict(
    millimeter=QtGui.QPageSize.Unit.Millimeter,
    point=QtGui.QPageSize.Unit.Point,
    inch=QtGui.QPageSize.Unit.Inch,
    pica=QtGui.QPageSize.Unit.Pica,
    didot=QtGui.QPageSize.Unit.Didot,
    cicero=QtGui.QPageSize.Unit.Cicero,
)

UnitStr = Literal["millimeter", "point", "inch", "pica", "didot", "cicero"]


class PageSize(QtGui.QPageSize):
    """Describes the size and name of a defined page size."""

    def __repr__(self):
        return get_repr(self, self.get_id())

    def __reduce__(self):
        return type(self), (self.id(),)

    def get_definition_units(self) -> UnitStr:
        """Get the definition unit.

        Returns:
            unit
        """
        units = self.definitionUnits()
        if units.value == -1:
            raise ValueError("Invalid page size")
        return UNITS.inverse[units]

    def get_id(self) -> str:
        """Get the standard page size id.

        Returns:
            page size id
        """
        return PAGE_SIZE_ID.inverse[self.id()]


if __name__ == "__main__":
    size = PageSize()
    print(repr(size))



from __future__ import annotations


try:  # pragma: no cover
    import re._constants as sre_constants
except ImportError:  # Python < 3.11
    import sre_constants  # type: ignore

import re

from prettyqt import core, gui
from prettyqt.utils import get_repr


class RegexPatternValidator(gui.Validator):
    """Validator which checks whether given string is a parseable regex pattern."""

    ID = "regex_pattern"
    error_occured = core.Signal(str)
    pattern_updated = core.Signal(object)

    def __repr__(self):
        return get_repr(self)

    def __eq__(self, other: object):
        return isinstance(other, type(self))

    def validate(  # type: ignore
        self, text: str, pos: int = 0
    ) -> tuple[gui.QValidator.State, str, int]:
        # if text == "":
        #     self.compiled = None
        #     return (self.Intermediate, text, pos)
        try:
            compiled = re.compile(text)
        except sre_constants.error as e:
            self.error_occured.emit(str(e))
            self.pattern_updated.emit(None)
            return self.State.Intermediate, text, pos
        else:
            self.error_occured.emit("")
            self.pattern_updated.emit(compiled)
            return self.State.Acceptable, text, pos


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    w = widgets.LineEdit()
    val = RegexPatternValidator()
    w.set_validator(val)
    w.show()
    app.exec()



from __future__ import annotations

from prettyqt.qt import QtCore
from prettyqt.utils import get_repr


class Point(QtCore.QPoint):
    """Defines a point in the plane using integer precision."""

    def __repr__(self):
        return get_repr(self, self.x(), self.y())

    @property
    def _x(self) -> int:
        return self.x()

    @property
    def _y(self) -> int:
        return self.y()

    __match_args__ = ("_x", "_y")

    def __getitem__(self, index) -> int:
        match index:
            case 0:
                return self.x()
            case 1:
                return self.y()
            case _:
                raise IndexError(f"Invalid index {index} for Point")

    def __setitem__(self, i, x: int):
        match i:
            case 0:
                return self.setX(x)
            case 1:
                return self.setY(x)
            case _:
                raise IndexError(f"Invalid index {i} for Point")

    def __reduce__(self):
        return type(self), (self.x(), self.y())


if __name__ == "__main__":
    p = Point(0, 1)
    match p:
        case Point(0, 1):
            raise Exception("yeah")



from __future__ import annotations

from collections.abc import Iterator, Mapping, MutableMapping, Sequence
import contextlib
import logging
import os
import sys

from typing import Any, Literal

from typing_extensions import Self

from prettyqt import core
from prettyqt.utils import bidict, datatypes


logger = logging.getLogger(__name__)

FormatStr = Literal["user", "system"]

FORMAT: bidict[FormatStr, core.QSettings.Format] = bidict(
    native=core.QSettings.Format.NativeFormat, ini=core.QSettings.Format.IniFormat
)

ScopeStr = Literal["user", "system"]

SCOPE: bidict[ScopeStr, core.QSettings.Scope] = bidict(
    user=core.QSettings.Scope.UserScope, system=core.QSettings.Scope.SystemScope
)


class Settings_(
    core.ObjectMixin, core.QSettings, MutableMapping, metaclass=datatypes.QABCMeta
):
    """Persistent platform-independent application settings."""

    # Setting class with original behavior, compatible with original QSettings
    def __init__(self, *args, settings_id: str | None = None):
        super().__init__(*args)
        self.settings_id = settings_id

    def __repr__(self):
        return f"{type(self).__name__}({dict(self)})"

    def __contains__(self, key: str) -> bool:
        return self.contains(key)

    def __enter__(self):
        if self.settings_id:
            self.beginGroup(self.settings_id)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.settings_id:
            self.endGroup()

    def __getitem__(self, index: str):
        if index not in self.allKeys():
            raise KeyError(index)
        return self.get_value(index)

    def __setitem__(self, name: str, value):
        return self.set_value(name, value)

    def __delitem__(self, key: str):
        if not self.contains(key):
            raise KeyError(key)
        return self.remove(key)

    def __iter__(self) -> Iterator[str]:
        return iter(self.allKeys())

    def __len__(self) -> int:
        return len(self.allKeys())

    def __ior__(self, other: Mapping):
        self.update(other)
        return self

    def set_value(self, key: str, value):
        if not self.applicationName():
            raise RuntimeError("no app name defined")
        if isinstance(value, Settings_):
            value = dict(value)
        self.setValue(key, value)

    def get_value(self, key: str, default=None):
        return self.value(key) if self.contains(key) else default

    @classmethod
    def set_default_format(cls, fmt: FormatStr | core.QSettings.Format):
        """Set the default format.

        Args:
            fmt: the default format to use
        """
        cls.setDefaultFormat(FORMAT.get_enum_value(fmt))

    @classmethod
    def get_default_format(cls) -> FormatStr:
        """Return default settings format.

        Returns:
            default settings format
        """
        return FORMAT.inverse[cls.defaultFormat()]

    def get_scope(self) -> ScopeStr:
        """Return scope.

        Returns:
            scope
        """
        return SCOPE.inverse[self.scope()]

    @classmethod
    def set_path(
        cls,
        fmt: FormatStr | core.QSettings.Format,
        scope: ScopeStr | core.QSettings.Scope,
        path: datatypes.PathType,
    ):
        """Set the path to the settings file.

        Args:
            fmt: the default format to use
            scope: the scope to use
            path: the path to set
        """
        fmt_val = FORMAT.get_enum_value(fmt)
        scope_val = SCOPE.get_enum_value(scope)
        cls.setPath(fmt_val, scope_val, os.fspath(path))

    @contextlib.contextmanager
    def edit_group(self, prefix: str):
        """Context manager for setting groups.

        Args:
            prefix: setting prefix for group
        """
        self.beginGroup(prefix)
        yield None
        self.endGroup()

    def write_array(self, prefix: str, array: Sequence) -> Iterator[tuple[Self, Any]]:
        """Context manager for writing arrays.

        Args:
            prefix: prefix for settings array
            array: array used to write settings

        Returns:
            Iterator yielding (setting, arrayitem) items
        """
        self.beginWriteArray(prefix, len(array))
        for i, _ in enumerate(array):
            self.setArrayIndex(i)
            yield (self, array[i])
        self.endArray()

    def read_array(self, prefix: str) -> Iterator[Settings_]:
        """Context manager for reading arrays.

        Args:
            prefix: prefix for settings array
        """
        size = self.beginReadArray(prefix)
        for i in range(size):
            self.setArrayIndex(i)
            yield self
        self.endArray()

    # MutableMapping overrides

    def keys(self) -> list[str]:
        return self.allKeys()

    def setdefault(self, key: str, default: Any = None) -> Any:
        if not self.contains(key):
            self.set_value(key, default)
            return default
        return self.get_value(key)

    @classmethod
    def register_extensions(
        cls,
        *exts: str,
        app_name: str | None = None,
        app_path: None | datatypes.PathType = None,
    ):
        logger.debug(f"assigning extensions {exts} to {app_name}")
        s = cls("HKEY_CURRENT_USER\\SOFTWARE\\Classes", Settings.Format.NativeFormat)
        if app_path is None:
            app_path = str(core.CoreApplication.get_application_file_path())
        app_path = os.fspath(app_path)
        if app_name is None:
            app_name = core.CoreApplication.applicationName()
        for ext in exts:
            s.setValue(f"{ext}/DefaultIcon/.", app_path)  # perhaps ,0 after app_path
            s.setValue(f"{ext}/.", app_name)
        if getattr(sys, "frozen", False):
            s.setValue(f"{app_name}/shell/open/command/.", f"{app_path} %1")
        else:
            command_value = rf'{sys.executable} -m {app_path} "%V"'
            s.setValue(f"{app_name}/shell/open/command/.", f"{command_value} %1")


class Settings(Settings_):
    """Settings class which wraps everything into a dict to preserve data types."""

    def set_value(self, key: str, value):
        match value:
            case Settings_():
                setting = dict(value=dict(value), typ="subsetting")
            case _:
                setting = dict(value=value, typ="regular")
        super().set_value(key, setting)

    def get_value(self, key: str, default=None):
        if not self.contains(key):
            return default
        val = self.value(key)
        # TODO: convert settings dicts back?
        match val:
            case {"value": setting}:
                return setting
            case _:
                # this is for migration
                self.set_value(key, val)
                return val


if __name__ == "__main__":
    settings = Settings("1", "2")
    settings["1"] = True
    settings |= dict(abc="abc")



from __future__ import annotations

from prettyqt import iconprovider
from prettyqt.qt import QtGui, QtWidgets


mod = QtWidgets.QStyle

MAPPING = {
    mod.StandardPixmap.SP_TitleBarMinButton: "mdi.window-minimize",
    mod.StandardPixmap.SP_TitleBarMenuButton: "mdi.microsoft-windows",
    mod.StandardPixmap.SP_TitleBarMaxButton: "mdi.window-maximize",
    mod.StandardPixmap.SP_TitleBarCloseButton: "mdi.window-close",
    mod.StandardPixmap.SP_TitleBarNormalButton: "mdi.window-restore",
    mod.StandardPixmap.SP_TitleBarShadeButton: "mdi.arrow-up",
    mod.StandardPixmap.SP_TitleBarUnshadeButton: "mdi.arrow-down",
    mod.StandardPixmap.SP_TitleBarContextHelpButton: "mdi.help",
    mod.StandardPixmap.SP_MessageBoxInformation: "mdi.information",
    mod.StandardPixmap.SP_MessageBoxWarning: "mdi.alert",
    mod.StandardPixmap.SP_MessageBoxCritical: "mdi.alert-circle-outline",
    mod.StandardPixmap.SP_MessageBoxQuestion: "mdi.tooltip-question-outline",
    mod.StandardPixmap.SP_DesktopIcon: "mdi.monitor",
    mod.StandardPixmap.SP_TrashIcon: "mdi.trash-can",
    mod.StandardPixmap.SP_ComputerIcon: "mdi.desktop-classic",
    mod.StandardPixmap.SP_DriveFDIcon: "mdi.floppy",
    mod.StandardPixmap.SP_DriveHDIcon: "mdi.harddisk",
    mod.StandardPixmap.SP_DriveCDIcon: "mdi.disc",
    mod.StandardPixmap.SP_DriveDVDIcon: "mdi.disc",
    mod.StandardPixmap.SP_DriveNetIcon: "mdi.network-outline",
    mod.StandardPixmap.SP_DirHomeIcon: "mdi.home-circle",
    mod.StandardPixmap.SP_DirOpenIcon: "mdi.folder-multiple-outline",
    mod.StandardPixmap.SP_DirClosedIcon: "mdi.folder-outline",
    mod.StandardPixmap.SP_DirIcon: "mdi.folder",
    mod.StandardPixmap.SP_DirLinkIcon: "mdi.link-box-outline",
    mod.StandardPixmap.SP_DirLinkOpenIcon: "mdi.link-plus",
    mod.StandardPixmap.SP_FileIcon: "mdi.file-outline",
    mod.StandardPixmap.SP_FileLinkIcon: "mdi.file-link-outline",
    mod.StandardPixmap.SP_FileDialogStart: "mdi.ray-start-arrow",
    mod.StandardPixmap.SP_FileDialogEnd: "mdi.ray-end-arrow",
    mod.StandardPixmap.SP_FileDialogToParent: "mdi.keyboard-return",
    mod.StandardPixmap.SP_FileDialogNewFolder: "mdi.folder-plus-outline",
    mod.StandardPixmap.SP_FileDialogDetailedView: "mdi.details",
    mod.StandardPixmap.SP_FileDialogInfoView: "mdi.information-outline",
    mod.StandardPixmap.SP_FileDialogContentsView: "mdi.content-paste",
    mod.StandardPixmap.SP_FileDialogListView: "mdi.format-list-bulleted",
    mod.StandardPixmap.SP_FileDialogBack: "mdi.arrow-left",
    mod.StandardPixmap.SP_DockWidgetCloseButton: "mdi.close-circle-outline",
    mod.StandardPixmap.SP_ToolBarHorizontalExtensionButton: "mdi.pan-horizontal",
    mod.StandardPixmap.SP_ToolBarVerticalExtensionButton: "mdi.pan-vertical",
    mod.StandardPixmap.SP_DialogOkButton: "mdi.check",
    mod.StandardPixmap.SP_DialogCancelButton: "mdi.cancel",
    mod.StandardPixmap.SP_DialogHelpButton: "mdi.help",
    mod.StandardPixmap.SP_DialogOpenButton: "mdi.play-protected-content",
    mod.StandardPixmap.SP_DialogSaveButton: "mdi.content-save-outline",
    mod.StandardPixmap.SP_DialogCloseButton: "mdi.window-close",
    mod.StandardPixmap.SP_DialogApplyButton: "mdi.thumb-up-outline",
    mod.StandardPixmap.SP_DialogResetButton: "mdi.lock-reset",
    mod.StandardPixmap.SP_DialogDiscardButton: "mdi.file-undo",
    mod.StandardPixmap.SP_DialogYesButton: "mdi.check",
    mod.StandardPixmap.SP_DialogNoButton: "mdi.cancel",
    mod.StandardPixmap.SP_ArrowUp: "mdi.arrow-up",
    mod.StandardPixmap.SP_ArrowDown: "mdi.arrow-down",
    mod.StandardPixmap.SP_ArrowLeft: "mdi.arrow-left",
    mod.StandardPixmap.SP_ArrowRight: "mdi.arrow-right",
    mod.StandardPixmap.SP_ArrowBack: "mdi.arrow-left",
    mod.StandardPixmap.SP_ArrowForward: "mdi.arrow-right",
    mod.StandardPixmap.SP_CommandLink: "mdi.apple-keyboard-command",
    mod.StandardPixmap.SP_VistaShield: "mdi.shield",
    mod.StandardPixmap.SP_BrowserReload: "mdi.reload",
    mod.StandardPixmap.SP_BrowserStop: "mdi.stop-circle-outline",
    mod.StandardPixmap.SP_MediaPlay: "mdi.play-circle-outline",
    mod.StandardPixmap.SP_MediaStop: "mdi.stop-circle-outline",
    mod.StandardPixmap.SP_MediaPause: "mdi.pause-circle-outline",
    mod.StandardPixmap.SP_MediaSkipForward: "mdi.skip-next-circle-outline",
    mod.StandardPixmap.SP_MediaSkipBackward: "mdi.skip-previous-circle-outline",
    mod.StandardPixmap.SP_MediaSeekForward: "mdi.step-forward",
    mod.StandardPixmap.SP_MediaSeekBackward: "mdi.step-backward",
    mod.StandardPixmap.SP_MediaVolume: "mdi.volume-high",
    mod.StandardPixmap.SP_MediaVolumeMuted: "mdi.volume-mute",
    mod.StandardPixmap.SP_LineEditClearButton: "mdi.close-circle-outline",
    mod.StandardPixmap.SP_DialogYesToAllButton: "mdi.check",
    mod.StandardPixmap.SP_DialogNoToAllButton: "mdi.cancel",
    mod.StandardPixmap.SP_DialogSaveAllButton: "mdi.content-save-outline",
    mod.StandardPixmap.SP_DialogAbortButton: "mdi.exit-to-app",
    mod.StandardPixmap.SP_DialogRetryButton: "mdi.repeat",
    mod.StandardPixmap.SP_DialogIgnoreButton: "mdi.minus-circle-outline",
    mod.StandardPixmap.SP_RestoreDefaultsButton: "mdi.history",
    mod.StandardPixmap.SP_TabCloseButton: "mdi.cancel",
    mod.StandardPixmap.SP_CustomBase: "mdi.firebase",
}


class MaterialIconStyle(QtWidgets.QProxyStyle):
    def standardIcon(
        self,
        standard_icon: QtWidgets.QStyle.StandardPixmap,
        option: QtWidgets.QStyleOption | None = None,
        widget: QtWidgets.QWidget | None = None,
    ) -> QtGui.QIcon:
        """Implement QProxyStyle.standardIcon."""
        icon_name = MAPPING.get(standard_icon)
        if icon_name is None:
            return super().standardIcon(standard_icon, option, widget)

        return iconprovider.get_icon(icon_name, as_qicon=True)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    style = MaterialIconStyle()
    app.setStyle(style)
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import bluetooth, core
from prettyqt.utils import bidict, get_repr


HostModeStr = Literal[
    "powered_off", "connectable", "discoverable", "discoverable_limited_inquiry"
]

HostMode = bluetooth.QBluetoothLocalDevice.HostMode
HOST_MODE: bidict[HostModeStr, HostMode] = bidict(
    powered_off=HostMode.HostPoweredOff,
    connectable=HostMode.HostConnectable,
    discoverable=HostMode.HostDiscoverable,
    discoverable_limited_inquiry=HostMode.HostDiscoverableLimitedInquiry,
)


Error = bluetooth.QBluetoothLocalDevice.Error

ErrorStr = Literal["none", "pairing", "missing_permissions", "unknown"]

ERROR: bidict[ErrorStr, Error] = bidict(
    none=Error.NoError,
    pairing=Error.PairingError,
    missing_permissions=Error.MissingPermissionsError,
    unknown=Error.UnknownError,
)


Pairing = bluetooth.QBluetoothLocalDevice.Pairing

PairingStr = Literal["unpaired", "paired", "authorized_paired"]

PAIRING: bidict[PairingStr, Pairing] = bidict(
    unpaired=Pairing.Unpaired,
    paired=Pairing.Paired,
    authorized_paired=Pairing.AuthorizedPaired,
)


class BluetoothLocalDevice(core.ObjectMixin, bluetooth.QBluetoothLocalDevice):
    """Enables access to the local Bluetooth device."""

    def __bool__(self):
        """Return True when local device is valid."""
        return self.isValid()

    def __repr__(self):
        return get_repr(self, self.address())

    def get_error(self) -> ErrorStr:
        """Get error code."""
        return ERROR.inverse[self.error()]

    def set_host_mode(self, mode: HostModeStr | HostMode):
        """Set host mode.

        Args:
            mode: host mode to use
        """
        self.setGridStyle(HOST_MODE.get_enum_value(mode))

    def get_host_mode(self) -> HostModeStr:
        """Return host mode.

        Returns:
            host mode
        """
        return HOST_MODE.inverse[self.gridStyle()]

    def get_pairing_status(
        self, address: bluetooth.QBluetoothAddress | int | str
    ) -> PairingStr:
        """Return pairing status.

        Arguments:
            address: bluetooth address

        Returns:
            pairing status
        """
        if isinstance(address, int | str):
            address = bluetooth.QBluetoothAddress(address)
        return PAIRING.inverse[self.pairingStatus(address)]

    def request_pairing(
        self,
        address: bluetooth.QBluetoothAddress | int | str,
        pairing: PairingStr | Pairing,
    ):
        """Request a pairing to given bluetooth address."""
        if isinstance(address, int | str):
            address = bluetooth.QBluetoothAddress(address)
        self.requestPairing(address, PAIRING.get_enum_value(pairing))

    def get_connected_devices(self) -> list[bluetooth.BluetoothAddress]:
        """Get addresses for connected devices."""
        return [bluetooth.BluetoothAddress(i) for i in self.connectedDevices()]

    @classmethod
    def for_all_devices(cls) -> list[bluetooth.BluetoothHostInfo]:
        """Get host info for all devices."""
        return [bluetooth.BluetoothHostInfo(i) for i in cls.allDevices()]


if __name__ == "__main__":
    devices = BluetoothLocalDevice.get_all_devices()



from __future__ import annotations

import contextlib

from typing import Literal

from prettyqt import gui
from prettyqt.utils import bidict


MOVE_MODE = bidict(
    move=gui.QTextCursor.MoveMode.MoveAnchor, keep=gui.QTextCursor.MoveMode.KeepAnchor
)

MoveModeStr = Literal["move", "keep"]

MOVE_OPERATION = bidict(
    no_move=gui.QTextCursor.MoveOperation.NoMove,
    start=gui.QTextCursor.MoveOperation.Start,
    start_of_line=gui.QTextCursor.MoveOperation.StartOfLine,
    start_of_block=gui.QTextCursor.MoveOperation.StartOfBlock,
    start_of_word=gui.QTextCursor.MoveOperation.StartOfWord,
    previous_block=gui.QTextCursor.MoveOperation.PreviousBlock,
    previous_char=gui.QTextCursor.MoveOperation.PreviousCharacter,
    previous_word=gui.QTextCursor.MoveOperation.PreviousWord,
    up=gui.QTextCursor.MoveOperation.Up,
    left=gui.QTextCursor.MoveOperation.Left,
    word_left=gui.QTextCursor.MoveOperation.WordLeft,
    end=gui.QTextCursor.MoveOperation.End,
    end_of_line=gui.QTextCursor.MoveOperation.EndOfLine,
    end_of_word=gui.QTextCursor.MoveOperation.EndOfWord,
    end_of_block=gui.QTextCursor.MoveOperation.EndOfBlock,
    next_block=gui.QTextCursor.MoveOperation.NextBlock,
    next_char=gui.QTextCursor.MoveOperation.NextCharacter,
    next_word=gui.QTextCursor.MoveOperation.NextWord,
    down=gui.QTextCursor.MoveOperation.Down,
    right=gui.QTextCursor.MoveOperation.Right,
    word_right=gui.QTextCursor.MoveOperation.WordRight,
    next_cell=gui.QTextCursor.MoveOperation.NextCell,
    previous_cell=gui.QTextCursor.MoveOperation.PreviousCell,
    next_row=gui.QTextCursor.MoveOperation.NextRow,
    previous_row=gui.QTextCursor.MoveOperation.PreviousRow,
)

MoveOperationStr = Literal[
    "no_move",
    "start",
    "start_of_line",
    "start_of_block",
    "start_of_word",
    "previous_block",
    "previous_char",
    "previous_word",
    "up",
    "left",
    "word_left",
    "end",
    "end_of_line",
    "end_of_word",
    "end_of_block",
    "next_block",
    "next_char",
    "next_word",
    "down",
    "right",
    "word_right",
    "next_cell",
    "previous_cell",
    "next_row",
    "previous_row",
]

SELECTION_TYPE = bidict(
    document=gui.QTextCursor.SelectionType.Document,
    block_under_cursor=gui.QTextCursor.SelectionType.BlockUnderCursor,
    line_under_cursor=gui.QTextCursor.SelectionType.LineUnderCursor,
    word_under_cursor=gui.QTextCursor.SelectionType.WordUnderCursor,
)

SelectionTypeStr = Literal[
    "document", "block_under_cursor", "line_under_cursor", "word_under_cursor"
]


class TextCursor(gui.QTextCursor):
    def __str__(self):
        return self.selectedText().replace("\u2029", "\n")

    def __contains__(self, other):
        return (
            self.selectionStart() <= other.selectionStart()
            and self.selectionEnd() >= other.selectionEnd()
        )

    def move_position(
        self, operation: MoveOperationStr, mode: MoveModeStr = "move", n: int = 1
    ) -> bool:
        if n < 0:
            raise ValueError(n)
        return self.movePosition(MOVE_OPERATION[operation], MOVE_MODE[mode], n)

    def set_position(self, pos: int | tuple[int, int], mode: MoveModeStr = "move"):
        """Set cursor to given position.

        0-indexed.

        Args:
            pos: Cursor position
            mode: Move mode
        """
        match pos:
            case int():
                self.setPosition(pos, MOVE_MODE[mode])
            case (int() as row, int() as col):
                position = self.document().find_block_by_number(row).position()
                position += col
                self.set_position(position, mode=mode)
            case _:
                raise TypeError(pos)

    def select(self, selection: SelectionTypeStr | gui.QTextCursor.SelectionType):
        if isinstance(selection, gui.QTextCursor.SelectionType):
            sel = selection
        else:
            sel = SELECTION_TYPE[selection]
        super().select(sel)

    def span(self) -> tuple[int, int]:
        return (self.anchor(), self.position())

    def get_cursor_position(self) -> tuple[int, int]:
        """Return the QTextCursor position.

        The position is a tuple made up of
        the line number (0 based) and the column number (0 based).
        :return: tuple(line, column)
        """
        return (self.blockNumber(), self.positionInBlock())

    def get_selection(self) -> gui.TextDocumentFragment:
        return gui.TextDocumentFragment(self.selection())

    def select_text(
        self,
        start_pos: int | tuple[int, int] | MoveOperationStr,
        end_pos: int | tuple[int, int] | MoveOperationStr,
    ) -> str:
        """Select text from start position to end position.

        Positions can be either an integer index or a move operation

        Args:
            start_pos: Start position
            end_pos: End position
        """
        match start_pos:
            case int() | (int(), int()):
                self.set_position(start_pos)
            case str():
                self.move_position(start_pos)
            case _:
                raise TypeError(start_pos)
        match end_pos:
            case int() | (int(), int()):
                self.set_position(end_pos, mode="keep")
            case str():
                self.move_position(end_pos, mode="keep")
            case _:
                raise TypeError(end_pos)
        return self.selectedText()

    def replace_text(
        self,
        start_pos: int,
        end_pos: MoveOperationStr | int,
        to_replace: str,
    ):
        self.set_position(start_pos)
        if isinstance(end_pos, int):
            self.set_position(end_pos, mode="keep")
        else:
            self.move_position(end_pos, mode="keep")
        self.insertText(to_replace)
        self.select_text(start_pos, start_pos + len(to_replace))

    @contextlib.contextmanager
    def edit_block(self, join_previous: bool = False):
        """Context manager for edit blocks. Can be used for undo actions."""
        self.joinPreviousEditBlock() if join_previous else self.beginEditBlock()
        yield
        self.endEditBlock()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PlainTextEdit()
    widget.show()
    app.exec()
    cursor = widget.selecter.get_text_cursor()
    print(cursor)



from __future__ import annotations

from prettyqt.qt import QtGui
from prettyqt.utils import get_repr


class PageRanges(QtGui.QPageRanges):
    """Represents a collection of page ranges."""

    def __bool__(self):
        return not self.isEmpty()

    def __getitem__(self, index: int):
        return self.get_range_list()[index]

    def __str__(self):
        return self.toString()

    def __repr__(self):
        return get_repr(self, self.get_range_list())

    def __contains__(self, other: int):
        return self.contains(other)

    def __add__(self, other: int | tuple[int, int]):
        if isinstance(other, int):
            self.addPage(other)
        else:
            self.addRange(*other)
        return self

    def __len__(self):
        return self.lastPage() - self.firstPage()

    def get_range_list(self) -> list[tuple[int, int]]:
        return [(i.from_, i.to) for i in self.toRangeList()]


if __name__ == "__main__":
    ranges = PageRanges()
    ranges.addPage(1)
    ranges.addPage(2)



from __future__ import annotations

import pathlib

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


mod = core.QLibraryInfo.LibraryPath

LocationStr = Literal[
    "prefix",
    "documentation",
    "headers",
    "libraries",
    "library_executables",
    "binaries",
    "plugins",
    "qml2_imports",
    "arch_data",
    "data",
    "translations",
    "examples",
    "tests",
    "settings",
]

LOCATION: bidict[LocationStr, mod] = bidict(
    prefix=mod.PrefixPath,
    documentation=mod.DocumentationPath,
    headers=mod.HeadersPath,
    libraries=mod.LibrariesPath,
    library_executables=mod.LibraryExecutablesPath,
    binaries=mod.BinariesPath,
    plugins=mod.PluginsPath,
    qml2_imports=mod.Qml2ImportsPath,
    arch_data=mod.ArchDataPath,
    data=mod.DataPath,
    translations=mod.TranslationsPath,
    examples=mod.ExamplesPath,
    tests=mod.TestsPath,
    settings=mod.SettingsPath,
)


class LibraryInfo(core.QLibraryInfo):
    """Information about the Qt library."""

    @classmethod
    def get_location(cls, location: LocationStr) -> pathlib.Path:
        path = cls.path(LOCATION.get_enum_value(location))
        return pathlib.Path(path)

    @classmethod
    def get_version(cls) -> core.VersionNumber:
        return core.VersionNumber(cls.version())


if __name__ == "__main__":
    path = LibraryInfo.get_location("prefix")



from __future__ import annotations

import functools

from typing import Literal, overload

from prettyqt import constants, gui, widgets


class MenuBar(widgets.WidgetMixin, widgets.QMenuBar):
    """Horizontal menu bar."""

    def __add__(self, other: gui.QAction | widgets.QMenu):
        self.add(other)
        return self

    def serialize_fields(self):
        return dict(
            default_up=self.isDefaultUp(),
            native_menu_bar=self.isNativeMenuBar(),
        )

    @overload
    def add_menu(self, menu_or_str: str) -> widgets.Menu:
        ...

    @overload
    def add_menu(self, menu_or_str: widgets.QMenu) -> gui.Action:
        ...

    @functools.singledispatchmethod
    def add_menu(self, title: str) -> widgets.Menu:
        action = gui.Action(parent=self, text=title)
        menu = widgets.Menu(title=title, parent=self)
        action.set_menu(menu)
        self.addAction(action)
        return menu

    @add_menu.register
    def _(self, menu: widgets.QMenu) -> gui.Action:
        action = gui.Action(parent=self, text=menu.title())
        action.set_menu(menu)
        self.addAction(action)
        return action

    def add_separator(self):
        self.addSeparator()

    def add(self, *items: widgets.QMenu | gui.QAction):
        for i in items:
            if isinstance(i, widgets.QMenu):
                action = gui.Action(self)
                action.set_text(i.title())
                action.set_menu(i)
                self.addAction(action)
            else:
                self.addAction(i)

    def set_corner_widget(
        self,
        widget: widgets.QWidget,
        corner: Literal["top_right", "top_left"] = "top_right",
    ):
        match corner:
            case "top_left":
                self.setCornerWidget(widget, constants.Corner.TopLeftCorner)
            case "top_right":
                self.setCornerWidget(widget, constants.Corner.TopRightCorner)
            case _:
                raise ValueError(corner)

    def get_corner_widget(
        self,
        corner: Literal["top_right", "top_left"] = "top_right",
    ) -> widgets.QWidget:
        match corner:
            case "top_left":
                return self.cornerWidget(constants.Corner.TopLeftCorner)
            case "top_right":
                return self.cornerWidget(constants.Corner.TopRightCorner)
            case _:
                raise ValueError(corner)


if __name__ == "__main__":
    app = widgets.app()
    win = widgets.MainWindow()
    menu_bar = MenuBar()
    menuaction = menu_bar.add_menu("test")
    act = menu_bar.add_action("action")
    sep = menu_bar.addSeparator()
    act2 = menu_bar.add_action("action2")
    menu = widgets.Menu("testaa")
    menu_bar.add(menu)
    win.setMenuBar(menu_bar)
    win.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import core
from prettyqt.utils import modelhelpers


logger = logging.getLogger(__name__)


MERMAID = """
    ``` mermaid
    classDiagram
      Shared_proxy <|-- Proxy_1_1
      Shared_proxy <|-- Proxy_2_1
      Proxy_1_1 <|-- Proxy_1_2
      Proxy_2_1 <|-- Proxy_2_2
      Root_model <-- Shared_proxy
      class Proxy_1_1{
      }
      class Proxy_2_1{
      }
      class Root_model{
      }
    ```
"""


class ProxyMapper(core.Object):
    """Class to map indexes / ItemSelections from one proxy to another.

    Also handles cases like:
    ```
                Root model
                    |
               shared proxy
               /          \
            Proxy_1_1     Proxy_2_1
              |            |
            Proxy_1_2     Proxy_2_2
    ```
    When mapping from 1_2 to 2_2, it will find the closest parent ("shared proxy" here),
    use mapToSource / mapSelectionFromSource until it gets there,
    and then mapFromSource / mapSelectionFromSource to get down to 2_2.
    ``` py
    mapper = ProxyMapper(proxy_1_2, proxy_2_1)
    index = proxy_1_2.index(0, 0)
    mapped_index = mapper.map_index(source=0, target=1, index)
    ```

    """

    def __init__(
        self,
        *proxies: core.QAbstractItemModel,
        **kwargs,
    ):
        super().__init__(**kwargs)
        chains = [modelhelpers.get_proxy_chain(proxy) for proxy in proxies]
        common_list = [
            element
            for element in chains[0]
            if all(element in sublist for sublist in chains[1:])
        ]
        if not common_list:
            raise RuntimeError("No common source model")
        common_source = common_list[0]
        logger.debug(f"Common source: {common_source}")
        self._chains = [chain[: chain.index(common_source)] for chain in chains]

    def map_index(
        self, source: int, target: int, index: core.ModelIndex
    ) -> core.ModelIndex:
        """Map index from source to target."""
        for model in self._chains[source]:
            logger.debug(f"mapping from {model!r} to {model.sourceModel()!r}")
            index = model.mapToSource(index)
        for model in reversed(self._chains[target]):
            logger.debug(f"mapping from {model.sourceModel()!r} to {model!r}")
            index = model.mapFromSource(index)
        return index

    def map_selection(
        self, source: int, target: int, selection: core.QItemSelection
    ) -> core.QItemSelection:
        """Map selection from source to target."""
        for model in self._chains[source]:
            logger.debug(f"mapping from {model!r} to {model.sourceModel()!r}")
            selection = model.mapSelectionToSource(selection)
        for model in reversed(self._chains[target]):
            logger.debug(f"mapping from {model.sourceModel()!r} to {model!r}")
            selection = model.mapSelectionFromSource(selection)
        return selection


if __name__ == "__main__":
    from prettyqt import gui, widgets

    app = widgets.app()
    data = dict(
        a=["abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedfaa"],
        b=[10000000, 2, 3, 4, 5, 6],
        c=[1, 2, 3, 4, 5, 6],
        d=[100000000, 2, 3, 4, 5, 6],
        e=[1000000, 2, 3, 4, 5, 6],
        f=[1000000, 2, 3, 4, 5, 6],
        g=[1, 2, 3, 4, 5, 6],
        h=[1, 2, 3, 4, 5, 6],
        i=[100000000000000, 2, 3, 4, 5, 6],
        j=[1, 2, 3, 4, 5, 6],
        k=[1, 2, 3, 4, 5, 6],
        jkfsj=[10, 20, 30, 40, 50, 60],
    )
    model = gui.StandardItemModel.from_dict(data)
    proxy = core.SortFilterProxyModel(object_name="common")
    proxy.setSourceModel(model)
    proxy_1_1 = core.SortFilterProxyModel(object_name="1_1")
    proxy_1_1.setSourceModel(proxy)
    proxy_1_2 = core.SortFilterProxyModel(object_name="1_2")
    proxy_1_2.setSourceModel(proxy_1_1)
    proxy_2_1 = core.SortFilterProxyModel(object_name="2_1")
    proxy_2_1.setSourceModel(proxy)
    proxy_2_2 = core.SortFilterProxyModel(object_name="2_2")
    proxy_2_2.setSourceModel(proxy_2_1)
    with app.debug_mode():
        mapper = ProxyMapper(proxy_1_2, proxy_2_2)
        index = mapper.map_index(0, 1, proxy_1_2.index(0, 0))
    print(index)



from __future__ import annotations

from prettyqt.qt import QtBluetooth
from prettyqt.utils import get_repr


class BluetoothAddress(QtBluetooth.QBluetoothAddress):
    """Assigns an address to the Bluetooth device."""

    def __repr__(self):
        return get_repr(self, self.toString())

    def __str__(self):
        return self.toString()

    def __bool__(self):
        """True when address is not null."""
        return not self.isNull()


if __name__ == "__main__":
    address = BluetoothAddress()



from __future__ import annotations

from collections.abc import Iterable, Iterator, Mapping
import functools

from typing import Any

from prettyqt import constants, core, iconprovider, widgets
from prettyqt.qt import QtGui
from prettyqt.utils import datatypes, listdelegators


class NoData:
    pass


class ListWidget(widgets.ListViewMixin, widgets.QListWidget):
    """Item-based list widget."""

    value_changed = core.Signal(object)

    def __init__(self, *args, selection_mode: str = "single", **kwargs):
        super().__init__(*args, selection_mode=selection_mode, **kwargs)
        self.itemSelectionChanged.connect(self.on_index_change)

    def __repr__(self):
        return f"{type(self).__name__}: {self.count()} items"

    def __getitem__(
        self, row: int | slice
    ) -> widgets.QListWidgetItem | listdelegators.ListDelegator[widgets.QListWidgetItem]:
        match row:
            case int():
                item = self.item(row)
                if item is None:
                    raise KeyError(row)
                return item
            case slice():
                count = self.itemCount() if row.stop is None else row.stop
                values = list(range(count)[row])
                ls = [self.item(i) for i in values]
                return listdelegators.ListDelegator(ls)
            case _:
                raise TypeError(row)

    def __delitem__(self, row: int):
        self.takeItem(row)

    def __add__(self, other: widgets.QListWidgetItem):
        self.addItem(other)
        return self

    def __iter__(self) -> Iterator[widgets.QListWidgetItem]:
        return iter(self.get_children())

    def __len__(self) -> int:
        return self.count()

    def sort(self, reverse: bool = False):
        order = constants.DESCENDING if reverse else constants.ASCENDING
        self.sortItems(order)

    def on_index_change(self):
        data = self.get_value()
        self.value_changed.emit(data)

    def get_children(self) -> listdelegators.ListDelegator[widgets.QListWidgetItem]:
        items = [self.item(row) for row in range(self.count())]
        return listdelegators.ListDelegator(items)

    def add_items(self, items: Iterable | Mapping):
        match items:
            case Mapping():
                for k, v in items.items():
                    self.add(v, k)
            case Iterable():
                for i in items:
                    if isinstance(i, tuple | list):
                        self.add(*i)
                    else:
                        self.add(i)
            case _:
                raise TypeError(items)

    @functools.singledispatchmethod
    def add_item(
        self,
        name: str = "",
        icon: datatypes.IconType = None,
        data: dict | None = None,
        foreground: QtGui.QBrush | None = None,
        background: QtGui.QBrush | None = None,
        font: QtGui.QFont | None = None,
        selected: bool = None,
        status_tip: str = "",
        tool_tip: str = "",
        whats_this: str | None = None,
        # text_alignment: Optional[str] = None,
        checkstate: constants.StateStr | constants.CheckState | None = None,
        flags: constants.ItemFlag | None = None,
        size_hint: datatypes.SizeType | None = None,
        is_user_type: bool = False,
    ) -> widgets.ListWidgetItem:
        typ = 1 if is_user_type else 0
        item = widgets.ListWidgetItem(name, self, typ)
        if icon is not None:
            icon = iconprovider.get_icon(icon)
            item.setIcon(icon)
        if data is not None:
            for k, v in data.items():
                item.set_data(k, v)
        if foreground is not None:
            item.setForeground(foreground)
        if background is not None:
            item.setBackground(background)
        if font is not None:
            item.setFont(font)
        if flags is not None:
            item.setFlags(flags)
        if selected:
            item.setSelected(selected)
        if status_tip:
            item.setStatusTip(status_tip)
        if tool_tip:
            item.setToolTip(tool_tip)
        if whats_this:
            item.setWhatsThis(whats_this)
        if size_hint is not None:
            item.set_size_hint(size_hint)
        if checkstate is not None:
            item.set_checkstate(checkstate)
        self.addItem(item)
        return item

    @add_item.register
    def _(self, listitem: widgets.QListWidgetItem):
        self.addItem(listitem)

    def add(self, label: str, data=NoData, icon: datatypes.IconType = None):
        if data is NoData:
            data = label
        item = widgets.ListWidgetItem(label)
        item.set_icon(icon)
        item.setData(constants.USER_ROLE, data)  # type: ignore
        self.addItem(item)

    def get_value(self) -> list[Any]:
        return [i.data(constants.USER_ROLE) for i in self.selectedItems()]  # type: ignore

    def set_value(self, value):
        for i in self.get_children():
            if i.data(constants.USER_ROLE) in value:  # type: ignore
                self.setCurrentItem(i)
                break

    def scroll_to_item(
        self,
        item: widgets.QListWidgetItem,
        mode: widgets.abstractitemview.ScrollHintStr
        | widgets.QAbstractItemView.ScrollHint = "ensure_visible",
    ):
        self.scrollToItem(item, widgets.abstractitemview.SCROLL_HINT.get_enum_value(mode))

    def find_items(
        self,
        text: str,
        column: int = 0,
        mode: constants.MatchFlagStr | constants.MatchFlag = "exact",
        recursive: bool = False,
        case_sensitive: bool = False,
    ) -> listdelegators.ListDelegator[widgets.QListWidgetItem]:
        flag = constants.MATCH_FLAGS.get_enum_value(mode)
        if recursive:
            flag |= constants.MatchFlag.MatchRecursive
        if case_sensitive:
            flag |= constants.MatchFlag.MatchCaseSensitive
        items = self.findItems(text, flag, column)  # type: ignore
        return listdelegators.ListDelegator(items)


if __name__ == "__main__":
    app = widgets.app()
    widget = ListWidget()
    for i in range(100):
        widget.add_item(str(i))
    widget.add("test", icon="mdi.timer")
    widget.add("test", icon="mdi.timer")
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets


class ErrorMessage(widgets.DialogMixin, widgets.QErrorMessage):
    """Error message display dialog."""


if __name__ == "__main__":
    app = widgets.app()
    widget = ErrorMessage()
    widget.set_icon("mdi.timer")
    widget.show()
    app.exec()



from __future__ import annotations

import logging

from prettyqt import constants, core
from prettyqt.qt import QtGui


logger = logging.getLogger(__name__)


class QObjectPropertiesModel(core.AbstractTableModel):
    HEADER = [
        "Property name",
        "Value",
        "Type",
        "User property",
        "Readable",
        "Writable",
        "Resettable",
        "Bindable",
        "Designable",
        "Constant",
        "Final",
        "Required",
        "Valid",
        "Stored",
        "Notifier",
        "User Type",
        # "Enumerator",
    ]
    SUPPORTS = core.QObject

    def __init__(self, qobject: core.QObject, **kwargs):
        self._qobject = qobject
        self._metaobj = core.MetaObject(self._qobject.metaObject())
        super().__init__(**kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, core.QObject)

    def set_qobject(self, qobject):
        with self.reset_model():
            self._qobject = qobject
            self._metaobj = core.MetaObject(self._qobject.metaObject())

    def columnCount(self, parent=None) -> int:
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE:
                return self.HEADER[section]
            case constants.VERTICAL, constants.DISPLAY_ROLE:
                prop = self._metaobj.get_property(section)
                return prop.propertyIndex()

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        prop = self._metaobj.get_property(index.row())
        match role, index.column():
            case constants.BACKGROUND_ROLE, _:
                return QtGui.QColor("lightblue") if prop.isUser() else None
            case constants.DISPLAY_ROLE, 0:
                return prop.name()
            case constants.FONT_ROLE, 0:
                font = QtGui.QFont()
                font.setBold(True)
                return font
            case constants.DISPLAY_ROLE | constants.EDIT_ROLE, 1:
                return prop.read(self._qobject)
            case constants.DISPLAY_ROLE, 2:
                return prop.get_meta_type().get_name()
            case constants.FONT_ROLE, 2:
                font = QtGui.QFont()
                font.setItalic(True)
                return font
            case constants.CHECKSTATE_ROLE, 3:
                return self.to_checkstate(prop.isUser())
            case constants.CHECKSTATE_ROLE, 4:
                return self.to_checkstate(prop.isReadable())
            case constants.CHECKSTATE_ROLE, 5:
                return self.to_checkstate(prop.isWritable())
            case constants.CHECKSTATE_ROLE, 6:
                return self.to_checkstate(prop.isResettable())
            case constants.CHECKSTATE_ROLE, 7:
                return self.to_checkstate(prop.isBindable())
            case constants.CHECKSTATE_ROLE, 8:
                return self.to_checkstate(prop.isDesignable())
            case constants.CHECKSTATE_ROLE, 9:
                return self.to_checkstate(prop.isConstant())
            case constants.CHECKSTATE_ROLE, 10:
                return self.to_checkstate(prop.isFinal())
            case constants.CHECKSTATE_ROLE, 11:
                return self.to_checkstate(prop.isRequired())
            case constants.CHECKSTATE_ROLE, 12:
                return self.to_checkstate(prop.isValid())
            case constants.CHECKSTATE_ROLE, 13:
                return self.to_checkstate(prop.isStored())
            case constants.DISPLAY_ROLE, 14:
                notifier = prop.get_notify_signal()
                return "" if notifier is None else notifier.get_name()
            case constants.DISPLAY_ROLE, 15:
                return prop.userType()
            # case constants.DISPLAY_ROLE, 8:
            #     enumerator = prop.get_enumerator()
            #     return "" if enumerator is None else enumerator.get_name()
            case constants.USER_ROLE, _:
                return prop.read(self._qobject)

    def setData(
        self,
        index: core.ModelIndex,
        value,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        prop = self._metaobj.get_property(index.row())
        match role, index.column():
            case constants.USER_ROLE | constants.EDIT_ROLE, _:
                prop.write(self._qobject, value)
                self.update_row(index.row())
                return True
        return False

    def rowCount(self, parent: core.QModelIndex | None = None) -> int:
        """Override for AbstractitemModel base method."""
        parent = parent or core.ModelIndex()
        if parent.column() > 0:
            return 0
        return 0 if parent.isValid() else self._metaobj.propertyCount()

    def flags(self, index: core.QModelIndex) -> constants.ItemFlag:
        prop = self._metaobj.get_property(index.row())
        if index.column() == 1 and prop.isWritable():
            return super().flags(index) | constants.IS_EDITABLE
        return super().flags(index)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    widget = debugging.example_widget()
    view = widgets.TableView()
    view.set_icon("mdi.folder")
    with app.debug_mode():
        model = QObjectPropertiesModel(widget)
        model.dataChanged.connect(view.repaint)
        view.set_model(model)
        view.set_selection_behavior("rows")
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor", column=1)
        view.show()
        widget.show()
        view.resize(500, 300)
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import widgets
from prettyqt.utils import bidict


BlurHintStr = Literal["performance", "quality", "animation"]

BLUR_HINTS: bidict[BlurHintStr, widgets.QGraphicsBlurEffect.BlurHint] = bidict(
    performance=widgets.QGraphicsBlurEffect.BlurHint.PerformanceHint,
    quality=widgets.QGraphicsBlurEffect.BlurHint.QualityHint,
    animation=widgets.QGraphicsBlurEffect.BlurHint.AnimationHint,
)


class GraphicsBlurEffect(widgets.GraphicsEffectMixin, widgets.QGraphicsBlurEffect):
    def set_blur_hints(self, *hints: BlurHintStr):
        if hints:
            flags = BLUR_HINTS.merge_flags(hints)
        else:
            flags = widgets.QGraphicsBlurEffect.BlurHint(0)  # type: ignore
        self.setBlurHints(flags)

    def get_blur_hints(self) -> list[BlurHintStr]:
        return BLUR_HINTS.get_list(self.blurHints())



from __future__ import annotations

import logging

from prettyqt import constants, core, custom_widgets, widgets


logger = logging.getLogger(__name__)

BOOL_ITEMS = {
    None: "Show all",
    constants.CheckState.Checked: "Show True",
    constants.CheckState.Unchecked: "Show False",
}


class FilterHeader(widgets.HeaderView):
    """A HeaderView subclass which includes widgets with filter possibilities.

    When setting the header view on an ItemView, a proxy model will be created which is
    linked to the filter widgets.
    The correct filter widget is automatically inferred from the content of the columns.

    So basically everything that needs to be done is the following:

    ```py
    model = MyModel()
    widget = widgets.TableView()
    widget.set_model(model)
    widget.h_header = custom_widgets.FilterHeader() # same as setHorizontalHeader()
    ```

    and you will get filter capabilities for your table.

    !!! note
        Since the FilterHeader will infer the column content type based on the first few
        rows, it will only work correctly for tables with homogenous data.

    <figure markdown>
      ![Image title](../../images/filterheader.png)
      <figcaption>FilterHeader widget</figcaption>
    </figure>

    """

    def __init__(self, parent: widgets.TableView):
        self._editors_visible = False
        self._editors = []
        self._proxy = parent.proxifier.get_proxy(
            "multi_column_filter",
            recursive_filtering_enabled=True,
        )
        self._padding = 6
        super().__init__(constants.HORIZONTAL, parent)
        self.setStretchLastSection(True)
        # self.setResizeMode(QHeaderView.Stretch)
        self.setDefaultAlignment(constants.ALIGN_CENTER_LEFT)
        # self.setSortIndicatorShown(False)
        self.sectionResized.connect(self._adjust_positions)
        parent.h_scrollbar.valueChanged.connect(self._adjust_positions)
        parent.model_changed.connect(self._update_filter_boxes)
        self.sectionResized.connect(self._adjust_positions)
        self._update_filter_boxes()
        self.update_geometries()

    @classmethod
    def setup_example(cls):
        w = widgets.TableView()
        widget = cls(parent=w)
        return widget

    def are_editors_visible(self) -> bool:
        return self._editors_visible

    def set_editors_visible(self, visible: bool):
        self._editors_visible = visible
        for editor in self._editors:
            editor.setVisible(visible)
        self.updateGeometries()

    def _update_filter_boxes(self):
        # TODO: deal with column changes by connecting to Model signals.
        # That way we wouldnt have to update all editors on change.
        while self._editors:
            editor = self._editors.pop()
            editor.deleteLater()
        self.create_editors()
        self._adjust_positions()

    def create_editors(self):
        # using parent model here bc we cant guarantee that we are already set to view.
        parent = self.parent()
        model = parent.model()
        self._proxy.clear_filters()
        for i in range(model.columnCount()):
            typ = model.get_column_type(i)
            if typ is bool:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val, constants.CHECKSTATE_ROLE)

                name = f"filter_combo_{i}"
                widget = widgets.ComboBox(margin=0, object_name=name, parent=self)
                widget.add_items(BOOL_ITEMS)
                widget.value_changed.connect(set_filter)
            elif typ in [int, float]:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val)

                name = f"filter_numwidget_{i}"
                widget = custom_widgets.NumFilterWidget(
                    margin=0, object_name=name, parent=self
                )
                widget.filter_changed.connect(set_filter)
                title = model.headerData(i, constants.HORIZONTAL, constants.DISPLAY_ROLE)
                widget.lineedit.setPlaceholderText(f"Filter {title}...")
            elif typ is str:

                def set_filter(val, i=i):
                    self._proxy.set_filter_value(i, val)

                name = f"filter_lineedit_{i}"
                widget = widgets.LineEdit(margin=0, object_name=name, parent=self)
                widget.value_changed.connect(set_filter)
                title = model.headerData(i, constants.HORIZONTAL, constants.DISPLAY_ROLE)
                widget.setPlaceholderText(f"Filter {title}...")
            else:
                widget = widgets.Widget()
            widget.show()
            self._editors.append(widget)

    def sizeHint(self) -> core.QSize:
        size = super().sizeHint()
        if self._editors:
            height = self._editors[0].sizeHint().height()
            size.setHeight(size.height() + height + self._padding)
        return size

    def updateGeometries(self):
        if self._editors:
            height = self._editors[0].sizeHint().height()
            self.setViewportMargins(0, 0, 0, height + self._padding)
        else:
            self.setViewportMargins(0, 0, 0, 0)
        super().updateGeometries()
        self._adjust_positions()

    def _adjust_positions(self):
        for index, editor in enumerate(self._editors):
            height = editor.sizeHint().height()
            compensate_y = 0
            compensate_x = 0
            match editor:
                case widgets.QComboBox():
                    compensate_y = +2
                case widgets.QPushButton():
                    compensate_y = -1
                case widgets.QCheckBox():
                    compensate_y = 4
                    compensate_x = 4
                case widgets.QWidget():
                    compensate_y = -1
            editor.move(
                self.sectionPosition(index) - self.offset() + 1 + compensate_x,
                height + (self._padding // 2) + compensate_y,
            )
            editor.resize(self.sectionSize(index), height)

    def set_filter_case_sensitive(self, value: bool):
        self._proxy.set_filter_case_sensitive(value)

    def clear_filters(self):
        for editor in self._editors:
            editor.clear()

    editors_visible = core.Property(
        bool,
        are_editors_visible,
        set_editors_visible,
        doc="Whether the filter widgets are visible",
    )


if __name__ == "__main__":
    from prettyqt import itemmodels

    app = widgets.app()
    with app.debug_mode():
        view = widgets.TableView()

        model = itemmodels.QObjectPropertiesModel(view, parent=view)
        # model = model.proxifier[:, 0:3]
        view.set_selection_behavior("rows")
        view.setEditTriggers(view.EditTrigger.AllEditTriggers)
        view.set_delegate("editor", column=1)

        view.setModel(model)
        view.resize(640, 480)
        view.set_selection_behavior("rows")
        view.adapt_sizes()
        header = FilterHeader(parent=view)
        view.setHorizontalHeader(header)
        # view.h_header._update_filter_boxes()
        view.h_header.visible_editors = True
        view.show()
        with app.debug_mode():
            app.exec()
            print(view.h_header._proxy._filters)



from __future__ import annotations

from collections.abc import Iterator
from typing import Literal, overload

from typing_extensions import Self

from prettyqt import constants, core, widgets
from prettyqt.utils import bidict, datatypes, listdelegators


# @contxtlib.contextmanager
# def get_sub_layout(
#     self,
#     sub_container: str,
#     parent: widgets.Widget | widgets.QLayout,
#     orientation: constants.OrientationStr | None = None,
#     stretch: int | None = None,
#     **kwargs,
# ):
#     match sub_container:
#         case "scroll":
#             scroller = widgets.ScrollArea(parent=parent)
#             scroller.setWidgetResizable(True)
#             widget = widgets.Widget(scroller)
#             scroller.set_widget(widget)
#             return widget.set_layout(orientation, **kwargs)
#         case "splitter":
#             return widgets.Splitter(orientation=orientation, parent=parent, **kwargs)
#         case "groupbox":
#             frame = widgets.GroupBox(parent=parent, **kwargs)
#             return frame.set_layout(orientation or "horizontal")
#         case "mainwindow":
#             return widgets.MainWindow(parent=parent)
#         case _:
#             from prettyqt import custom_widgets
#             CONTEXT_LAYOUTS = dict(
#                 horizontal=widgets.HBoxLayout,
#                 vertical=widgets.VBoxLayout,
#                 grid=widgets.GridLayout,
#                 form=widgets.FormLayout,
#                 stacked=widgets.StackedLayout,
#                 flow=custom_widgets.FlowLayout,
#                 border=custom_widgets.BorderLayout,
#             )
#             Klass = CONTEXT_LAYOUTS[sub_container]
#             layout = Klass(**kwargs)
#             widget = widgets.Widget(parent=parent)
#             widget.set_layout(layout)
#             return layout


SIZE_CONSTRAINT = bidict(
    default=widgets.QLayout.SizeConstraint.SetDefaultConstraint,
    fixed=widgets.QLayout.SizeConstraint.SetFixedSize,
    minimum=widgets.QLayout.SizeConstraint.SetMinimumSize,
    maximum=widgets.QLayout.SizeConstraint.SetMaximumSize,
    min_and_max=widgets.QLayout.SizeConstraint.SetMinAndMaxSize,
    none=widgets.QLayout.SizeConstraint.SetNoConstraint,
)

SizeConstraintStr = Literal[
    "default", "fixed", "minimum", "maximum", "min_and_max", "none"
]

LayoutTypeStr = Literal["horizontal", "vertical", "grid", "form", "stacked", "flow"]


class LayoutMixin(core.ObjectMixin, widgets.LayoutItemMixin):
    def __init__(self, *args, margin=None, **kwargs):
        self._next_container = None
        self._stack = []
        super().__init__(*args, **kwargs)
        if margin is not None:
            self.set_margin(margin)

    @overload
    def __getitem__(
        self, index: slice
    ) -> listdelegators.ListDelegator[widgets.QWidget | widgets.QLayout]:
        ...

    @overload
    def __getitem__(self, index: int | str) -> widgets.QWidget | widgets.QLayout:
        ...

    def __getitem__(
        self, index: str | int | slice
    ) -> (
        widgets.QWidget
        | widgets.QLayout
        | listdelegators.ListDelegator[widgets.QWidget | widgets.QLayout]
    ):
        match index:
            case int():
                if index < 0:
                    index + self.count()
                if index < 0 or index >= self.count():
                    raise IndexError(index)
                item = self.itemAt(index)
                return i if (i := item.widget()) is not None else item.layout()
            case str():
                if (item := self.find_child(typ=core.QObject, name=index)) is not None:
                    return item
                raise KeyError(index)
            case slice():
                stop = index.stop or self.count()
                rng = range(index.start or 0, stop, index.step or 1)
                widgets = [self[i] for i in rng]
                return listdelegators.ListDelegator(widgets, parent=self)
            case _:
                raise TypeError(index)

    def __setitem__(self, key, value):
        if self._container != self:
            self._container.__setitem__(key, value)

    def __delitem__(self, item: int | widgets.QLayoutItem):
        if isinstance(item, int):
            item = self.itemAt(item)
        self.removeItem(item)
        item.deleteLater()

    def __len__(self) -> int:
        return self.count()

    def __iter__(self) -> Iterator[widgets.QWidget | widgets.QLayout | None]:
        return iter(self[i] for i in range(self.count()))

    def __contains__(self, item: widgets.QWidget | widgets.QLayoutItem):
        return self.indexOf(item) >= 0

    def __iadd__(self, item, *args, **kwargs):
        self.add(item, *args, **kwargs)
        return self

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"sizeConstraint": SIZE_CONSTRAINT}
        return maps

    def add(self, item, *args, **kwargs):
        match item:
            case widgets.QWidget():
                self._container.addWidget(item, *args, **kwargs)
            case widgets.QLayout():
                self._container.addLayout(item, *args, **kwargs)
            case widgets.QLayoutItem():
                self._container.addItem(item, *args, **kwargs)
            case list():
                for i in item:
                    self._container.add(i, *args, **kwargs)
        return item

    def __enter__(self):
        def enter(item):
            if item._next_container is not None:
                enter(item._next_container)
                item._stack.append(item._next_container)
                item._next_container = None
            return item

        return enter(self)

    def __exit__(self, *_):
        def exit(item):
            if item._stack:
                item = item._stack.pop()
                exit(item)

        exit(self)

    @property
    def _container(self):
        return self._stack[-1] if self._stack else self

    def get_sub_layout(
        self,
        layout: str,
        orientation: constants.OrientationStr | None = None,
        stretch: int | None = None,
        **kwargs,
    ) -> Self:
        from prettyqt import custom_widgets

        CONTEXT_LAYOUTS = dict(
            horizontal=widgets.HBoxLayout,
            vertical=widgets.VBoxLayout,
            grid=widgets.GridLayout,
            form=widgets.FormLayout,
            stacked=widgets.StackedLayout,
            flow=custom_widgets.FlowLayout,
            splitter=widgets.Splitter,
            scroll=widgets.ScrollArea,
            frame=widgets.GroupBox,
        )
        Klass = CONTEXT_LAYOUTS[layout]
        match self._container:
            case widgets.QWidget() if layout == "scroll":
                scroller = Klass(parent=self._container)
                scroller.setWidgetResizable(True)
                widget = widgets.Widget()
                scroller.set_widget(widget)
                new = widget.set_layout(orientation, **kwargs)
            case widgets.QLayout() if layout == "scroll":
                scroller = Klass(parent=self._container)
                scroller.setWidgetResizable(True)
                widget = widgets.Widget()
                scroller.set_widget(widget)
                new = widget.set_layout(orientation, **kwargs)
                self._container.add(new)
            case widgets.QWidget() if layout == "splitter":
                new = Klass(orientation=orientation, parent=self._container, **kwargs)
            case widgets.QLayout() if layout == "splitter":
                new = Klass(orientation=orientation, **kwargs)
                self._container.add(new)
            case widgets.QWidget() if layout == "frame":
                frame = Klass(parent=self._container, **kwargs)
                widget = widgets.Widget()
                new = widget.set_layout(orientation or "horizontal")
                frame.set_layout(new)
            case widgets.QLayout() if layout == "frame":
                frame = Klass(**kwargs)
                widget = widgets.Widget()
                new = widget.set_layout(orientation or "horizontal")
                frame.set_layout(new)
                self._container.add(new)
            case widgets.QMainWindow():
                widget = widgets.Widget(parent=self._container)
                self._container.setCentralWidget(widget)
                new = Klass(widget, **kwargs)
            case widgets.QScrollArea():
                widget = widgets.Widget(parent=self._container)
                self._container.setWidget(widget)
                self._container.setWidgetResizable(True)
                new = widget.set_layout("vertical", **kwargs)
            case widgets.QSplitter():
                widget = widgets.Widget(parent=self._container)
                self._container.addWidget(widget)
                new = Klass(widget, **kwargs)
            case None | widgets.QWidget():
                new = Klass(self._container, **kwargs)
            case widgets.QLayout():
                new = Klass(**kwargs)
                if stretch:
                    self._container.add(new, stretch)
                else:
                    self._container.add(new)
        new._stack = []
        new._next_container = None
        self._next_container = new
        return self

    def item_at(self, pos_or_index: int | core.Point) -> widgets.QLayoutItem:
        match pos_or_index:
            case int():
                return super().itemAt(pos_or_index)
            case core.QPoint():
                for i in range(self.count()):
                    item = self.itemAt(i)
                    if item.geometry().contains(pos_or_index):
                        return item
            case _:
                raise ValueError(pos_or_index)

    def clear(self):
        for i in reversed(range(self.count())):
            self.takeAt(i)

    # def takeAt(self, index: int):
    #     if index < 0:
    #         index = self.count() + index
    #     return super().takeAt(index)

    def get_items(self):
        return [self.itemAt(i) for i in range(self.count())]

    def get_children(
        self,
    ) -> listdelegators.ListDelegator[widgets.QWidget | widgets.QLayout]:
        return listdelegators.ListDelegator(self)

    def set_margin(self, margin: datatypes.MarginsType | None):
        match margin:
            case None:
                self.unsetContentsMargins()
            case _:
                margin = datatypes.to_margins(margin)
                self.setContentsMargins(margin)

    def set_spacing(self, pixels: int):
        self.setSpacing(pixels)

    def set_size_constraint(
        self, mode: SizeConstraintStr | widgets.QLayout.SizeConstraint
    ):
        """Set the size mode of the layout.

        Args:
            mode: size mode for the layout
        """
        self.setSizeConstraint(SIZE_CONSTRAINT.get_enum_value(mode))

    def get_size_constraint(self) -> SizeConstraintStr:
        """Return current size mode.

        Returns:
            size mode
        """
        return SIZE_CONSTRAINT.inverse[self.sizeConstraint()]

    def set_alignment(
        self,
        alignment: constants.AlignmentStr | constants.AlignmentFlag,
        item: widgets.QWidget | widgets.QLayout | None = None,
    ) -> bool:
        """Set the alignment for widget / layout to alignment.

        Returns true if w is found in this layout (not including child layouts).

        Args:
            alignment: alignment for the layout
            item: set alignment for specific child only
        """
        if item is not None:
            return self.setAlignment(item, constants.ALIGNMENTS.get_enum_value(alignment))
        else:
            return self.setAlignment(constants.ALIGNMENTS.get_enum_value(alignment))

    # def add(self, *items: widgets.QWidget | widgets.QLayout):
    #     for i in items:
    #         match i:
    #             case widgets.QWidget():
    #                 self.addWidget(i)
    #             case widgets.QLayout():
    #                 w = widgets.Widget()
    #                 w.set_layout(i)
    #                 self.addWidget(w)
    #             case _:
    #                 raise TypeError("add_item only supports widgets and layouts")


class Layout(LayoutMixin, widgets.QLayout):
    """The base class of geometry managers."""


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()
    widget = debugging.example_widget()

    widget.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from prettyqt import core
from prettyqt.utils import get_repr


class ValueFilterProxyModel(core.SortFilterProxyModel):
    """Proxy model for filtering based on non-str values.

    Sometimes it is needed to filter for non-str values, especially when it is required
    to filter based on a different role than DisplayRole.

    Same as the Qt QSortFilterProxyModel, this proxy respects the filterRole and
    filterKeyColumn properties.
    """

    ID = "value_filter"

    def __init__(self, filter_value=None, **kwargs):
        self._filter_value = filter_value
        super().__init__(**kwargs)

    def __repr__(self):
        return get_repr(self, self._filter_value)

    def filterAcceptsRow(self, source_row: int, parent: core.ModelIndex) -> bool:
        if self._filter_value is None:
            return True
        column = self.filterKeyColumn()
        role = self.filterRole()
        source_model = self.sourceModel()
        idx = source_model.index(source_row, column, parent)
        value = source_model.data(idx, role)
        return value == self._filter_value

    def set_filter_value(self, value):
        """Set the filter value."""
        self._filter_value = value
        self.invalidateRowsFilter()

    def get_filter_value(self):
        """Get the filter value."""
        return self._filter_value

    filter_value = core.Property(
        object,
        get_filter_value,
        set_filter_value,
        user=True,
        doc="Value to use for filtering",
    )


if __name__ == "__main__":
    from prettyqt import constants, widgets

    app = widgets.app()
    proxy = ValueFilterProxyModel(True, filter_role=constants.CHECKSTATE_ROLE)



"""Provides classes / tools for creating and embedding state machines from SCXML files."""

from __future__ import annotations

from prettyqt.qt.QtScxml import *  # noqa: F403

from .scxmlcompiler import ScxmlCompiler
from scxmlstatemachine import ScxmlStateMachine
from .scxmldatamodel import ScXmlDataModel
from .scxmlcppdatamodel import ScXmlCppDataModel
from .scxmlnulldatamodel import ScXmlNullDataModel
from .scxmlinvokableservice import ScXmlInvokableService
from .scxmlinvokableservicefactory import ScXmlInvokableServiceFactory
from prettyqt.qt import QtScxml

QT_MODULE = QtScxml

__all__ = [
    "ScxmlCompiler",
    "ScxmlStateMachine",
    "ScXmlDataModel",
    "ScXmlCppDataModel",
    "ScXmlNullDataModel",
    "ScXmlInvokableService",
    "ScXmlInvokableServiceFactory",
]



from __future__ import annotations

from collections.abc import Iterable, Mapping, MutableMapping
import hashlib
import logging

from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict


ResizeModeStr = Literal["interactive", "fixed", "stretch", "resize_to_contents"]

RESIZE_MODE: bidict[ResizeModeStr, widgets.QHeaderView.ResizeMode] = bidict(
    interactive=widgets.QHeaderView.ResizeMode.Interactive,
    fixed=widgets.QHeaderView.ResizeMode.Fixed,
    stretch=widgets.QHeaderView.ResizeMode.Stretch,
    resize_to_contents=widgets.QHeaderView.ResizeMode.ResizeToContents,
)

logger = logging.getLogger(__name__)


class HeaderWrapper:
    def __init__(self, indexer, widget: widgets.QHeaderView):
        self._widget = widget
        self._range = (
            range(
                indexer.start or 0,
                indexer.stop or self._widget.count(),
                indexer.step or 1,
            )
            if isinstance(indexer, slice)
            else range(indexer, indexer + 1)
        )

    def set_hidden(self, hide: bool = True):
        for i in self._range:
            self._widget.setSectionHidden(i, hide)

    def resize(self, size: int | Iterable[int]):
        match size:
            case int():
                for i in self._range:
                    self._widget.resizeSection(i, size)
            case Iterable():
                for i, j in zip(self._range, size):
                    self._widget.resizeSection(i, j)
            case _:
                raise TypeError(size)

    def set_resize_mode(self, mode: ResizeModeStr):
        for i in self._range:
            self._widget.setSectionResizeMode(i, RESIZE_MODE[mode])

    def set_sort_indicator(self, ascending: bool = True):
        sort_order = constants.ASCENDING if ascending else constants.DESCENDING
        for i in self._range:
            self._widget.setSortIndicator(i, sort_order)

    def is_in_visual_range(self) -> bool:
        return self._widget.is_in_visual_range(
            self._range.start
        ) and self._widget.is_in_visual_range(self._range.stop)


class HeaderViewMixin(widgets.AbstractItemViewMixin):
    section_visiblity_changed = core.Signal(int, bool)
    section_resized_by_user = core.Signal(int, int, int)

    def __init__(
        self,
        orientation: constants.OrientationStr | constants.Orientation,
        parent: widgets.QWidget | None = None,
        **kwargs,
    ):
        ori = constants.ORIENTATION.get_enum_value(orientation)
        super().__init__(ori, parent=parent, **kwargs)
        if ori == constants.HORIZONTAL:
            self.installEventFilter(self)

        self.setSectionsMovable(True)
        self.setSectionsClickable(True)
        self.sectionResized.connect(self._on_section_resize)
        self._handle_section_is_pressed = False
        self.setResizeContentsPrecision(100)
        self._widget_name = parent.objectName() if parent is not None else ""

    def __getitem__(self, index):
        match index:
            case int() if index >= self.count() or index < 0:
                raise IndexError(index)
            case int():
                return HeaderWrapper(index, self)
            case str():
                index = self.get_section_for_label(index)
                return HeaderWrapper(index, self)
            case _:
                raise TypeError(index)

    @classmethod
    def setup_example(cls):
        model = core.StringListModel(["a", "b", "c"])
        header = cls(orientation="vertical")
        header.setModel(model)
        return header

    def _get_map(self):
        maps = super()._get_map()
        maps |= {"defaultAlignment": constants.ALIGNMENTS}
        return maps

    def set_sort_indicator(self, logical_index: int | None, ascending: bool = True):
        idx = -1 if logical_index is None else logical_index
        sort_order = constants.ASCENDING if ascending else constants.DESCENDING
        self.setSortIndicator(idx, sort_order)

    def set_sections_movable(self, value: bool, include_first: bool = False):
        self.setSectionsMovable(value)
        self.setFirstSectionMovable(include_first)

    def mousePressEvent(self, e):
        super().mousePressEvent(e)
        shape = self.get_cursor().get_shape()
        self._handle_section_is_pressed = shape == "split_horizontal"

    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        self._handle_section_is_pressed = False

    def _on_section_resize(self, logical_index, old_size, new_size):
        if self._handle_section_is_pressed:
            self.section_resized_by_user.emit(logical_index, old_size, new_size)

    def generate_header_id(self):
        # return f"{self._widget_name}.state"
        column_names = ",".join(self.get_section_labels())
        columns_hash = hashlib.md5(column_names.encode()).hexdigest()
        return f"{type(self).__name__}_{columns_hash}.state"

    def save_state(self, settings: MutableMapping | None = None, key: str | None = None):
        settings = core.Settings() if settings is None else settings
        key = self.generate_header_id() if key is None else key
        settings[key] = self.saveState()

    def load_state(self, settings: Mapping | None = None, key: str | None = None) -> bool:
        settings = core.Settings() if settings is None else settings
        key = self.generate_header_id() if key is None else key
        if (state := settings.get(key)) is not None:
            if isinstance(state, str):
                state = state.encode()
            self.restoreState(state)
            return True
        return False

    def resize_sections(self, mode: ResizeModeStr = "resize_to_contents"):
        self.resizeSections(RESIZE_MODE[mode])

    def get_resize_mode(self, col: int) -> ResizeModeStr:
        val = self.sectionResizeMode(col)
        return RESIZE_MODE.inverse[val]

    def set_resize_mode(
        self,
        mode: ResizeModeStr | widgets.QHeaderView.ResizeMode,
        precision: int | None = None,
        cascading: bool | None = None,
        stretch_last_section: bool | None = None,
        default_section_size: int | None = None,
        maximum_section_size: int | None = None,
        minimum_section_size: int | None = None,
    ):
        if stretch_last_section is not None:
            self.stretchLastSection(stretch_last_section)
        if minimum_section_size is not None:
            self.setMinimumSectionSize(minimum_section_size)
        if default_section_size is not None:
            self.setDefaultSectionSize(default_section_size)
        if maximum_section_size is not None:
            self.setMaximumSectionSize(maximum_section_size)
        if precision is not None:
            self.setResizeContentsPrecision(precision)
        if cascading is not None:
            self.setCascadingSectionResizes(cascading)
        self.setSectionResizeMode(RESIZE_MODE.get_enum_value(mode))

    def get_section_labels(self) -> list[str]:
        """Return all section labels as a list."""
        model = self.model()
        orientation = self.orientation()
        return [
            model.headerData(i, orientation, constants.DISPLAY_ROLE)
            for i in range(self.count())
        ]

    def is_in_visual_range(self, section: int) -> bool:
        if self.orientation() == constants.VERTICAL:
            length = self.height()
            count = self.model().rowCount()
        else:
            length = self.width()
            count = self.model().columnCount()
        if count == 0 or section >= count:
            return False
        start = max(self.visualIndexAt(0), 0)
        end = self.visualIndexAt(length)
        if end == -1:
            end = count - 1
        sections = (self.logicalIndex(i) for i in range(start, end))
        return section in sections

    def get_section_for_label(self, label) -> int:
        """Return index of first section with given label."""
        model = self.model()
        orientation = self.orientation()
        for i in range(self.count()):
            if model.headerData(i, orientation, constants.DISPLAY_ROLE) == label:
                return i
        raise ValueError(label)

    def eventFilter(self, source, event):
        if event.type() == core.Event.Type.ContextMenu:
            menu = self.createPopupMenu()
            menu.exec(self.mapToGlobal(event.pos()))
            logger.info(f"KeypressEvent {event} taken by EventFilter")
            return True
        return False

    def get_header_actions(self) -> list[gui.Action]:
        menu = self.createPopupMenu()
        return menu.actions()

    def createPopupMenu(self) -> widgets.Menu:
        menu = widgets.Menu(parent=self)
        actions = []
        labels = self.get_section_labels()[1:]
        for i, header_label in enumerate(labels, start=1):
            val = not self.isSectionHidden(i)
            action = gui.Action(
                text=str(header_label),
                checkable=True,
                checked=val,
                triggered=self._on_toggle_action_triggered,
            )
            action.setData(i)
            actions.append(action)
        menu.add_actions(actions)
        return menu

    def _on_toggle_action_triggered(self):
        action = self.sender()
        section = action.data()
        self.toggle_section_visibility(section)

    def toggle_section_visibility(self, logical_index: int):
        self.setSectionHidden(logical_index, not self.isSectionHidden(logical_index))
        logger.info(f"toggling {logical_index}")

    def setSectionHidden(self, logical_index: int, hide: bool):
        self.section_visiblity_changed.emit(logical_index, hide)
        super().setSectionHidden(logical_index, hide)

    def set_sizes(self, sizes: Iterable[int | None]):
        for i, size in enumerate(sizes):
            if size is not None:
                self.resizeSection(i, size)

    def set_default_section_size(self, size: int | None):
        if size is None:
            self.resetDefaultSectionSize()
        else:
            self.setDefaultSectionSize(size)

    def stretch_last_section(self, stretch: bool = True):
        self.setStretchLastSection(stretch)

    def get_default_alignment(self) -> constants.AlignmentStr:
        return constants.ALIGNMENTS.inverse[self.defaultAlignment()]

    def get_orientation(self) -> constants.OrientationStr:
        """Return current orientation.

        Returns:
            orientation
        """
        return constants.ORIENTATION.inverse[self.orientation()]


class HeaderView(HeaderViewMixin, widgets.QHeaderView):
    """Header row or header column for item views."""


if __name__ == "__main__":
    from prettyqt import debugging

    app = widgets.app()
    table = debugging.example_tree()
    table.h_header[1:].set_hidden(True)
    table.show()

    with app.debug_mode():
        app.exec()



"""Provides QtPositioning classes and functions."""

from prettyqt.qt import PYQT6, PYSIDE6


if PYSIDE6:
    from PySide6.QtPositioning import *
elif PYQT6:
    from PyQt6.QtPositioning import *
else:
    raise ModuleNotFoundError("No Qt bindings could be found")



"""This module contains the editor widget implementation."""

from __future__ import annotations

from prettyqt import widgets
from prettyqt.custom_widgets import regexeditor


def run():
    app = widgets.app()
    widget = regexeditor.RegexEditorWidget()
    widget.show()
    app.exec()


if __name__ == "__main__":
    app = widgets.app()
    widget = regexeditor.RegexEditorWidget()
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, gui, widgets


class BaseMarkupDelegate(widgets.StyledItemDelegate):
    # An alternative approach would be go grab a pixmap from a QLabel for painting.
    # (see ButtonDelegate)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.doc = gui.TextDocument()
        self.text_option = gui.TextOption()
        self.text_option.setWrapMode(gui.TextOption.WrapMode.NoWrap)
        self.doc.setDefaultTextOption(self.text_option)
        self.doc.setDocumentMargin(0)

    def paint(
        self,
        painter: gui.QPainter,
        option: widgets.QStyleOptionViewItem,
        index: core.ModelIndex,
    ):
        painter.save()
        option = widgets.StyleOptionViewItem(option)
        self.initStyleOption(option, index)
        self.prepare_doc(option)
        # draw everything without text
        option.text = ""
        option.widget.style().drawControl(
            widgets.QStyle.ControlElement.CE_ItemViewItem, option, painter
        )

        # now we find position of our label and display our TextDocument there.
        icon_size = option.icon.actualSize(option.decorationSize)
        margin = icon_size.width()
        painter.translate(option.rect.left() + margin, option.rect.top())
        clip = core.QRectF(0, 0, option.rect.width() + margin, option.rect.height())
        self.doc.drawContents(painter, clip)

        painter.restore()

    def sizeHint(self, option: widgets.QStyleOptionViewItem, index: core.ModelIndex):
        option = widgets.StyleOptionViewItem(option)
        self.initStyleOption(option, index)
        self.prepare_doc(option)
        return core.QSize(int(self.doc.idealWidth()), int(self.doc.size().height()))

    def prepare_doc(self, option: widgets.QStyleOptionViewItem):
        self.text_option.setAlignment(option.displayAlignment)
        self.doc.setDefaultFont(option.font)
        self.set_text(option.text)
        # self.doc.setTextWidth(option.rect.width())
        self.doc.setPageSize(core.QSizeF(option.rect.width(), option.rect.height()))


class HtmlItemDelegate(BaseMarkupDelegate):
    """Delegate do display HTML text."""

    ID = "html"

    def set_text(self, text: str):
        self.doc.setHtml(text)


class MarkdownItemDelegate(BaseMarkupDelegate):
    """Delegate do display Markdown text."""

    ID = "markdown"

    def __init__(self, *args, mode: str = "github", **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = mode

    def set_text(self, text: str):
        Feat = self.doc.MarkdownFeature
        match self._mode:
            case "no_html":
                self.doc.setMarkdown(text, Feat.MarkdownNoHTML)
            case "commonmark":
                self.doc.setMarkdown(text, Feat.MarkdownDialectCommonMark)
            case "github" | "":
                self.doc.setMarkdown(text, Feat.MarkdownDialectGitHub)
            case _:
                raise ValueError(self._mode)

    def get_markup_mode(self, str):
        return self._mode

    def set_markup_mode(self, mode: str):
        self._mode = mode

    markup_mode = core.Property(
        str,
        get_markup_mode,
        set_markup_mode,
        doc="Which markdown dialect to use.",
    )


if __name__ == "__main__":
    from prettyqt import iconprovider

    app = widgets.app()
    model = gui.StandardItemModel()
    model.add("test")
    app = widgets.app()
    w = widgets.ListView()
    w.set_delegate(HtmlItemDelegate())
    w.set_model(model)
    item = gui.StandardItem("<b>test</b>")
    item.setBackground(gui.Color("green"))
    item.setIcon(iconprovider.get_icon("mdi.folder", as_qicon=True))
    item.setFont(gui.Font.mono())
    item.set_data(iconprovider.get_icon("mdi.folder"), "user")
    model += item
    w.show()
    app.exec()



from __future__ import annotations

import logging

from typing import Any

from prettyqt import constants, core, itemmodels


logger = logging.getLogger(__name__)


class SliceCheckableProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to make a model checkable.

    ### Example

    ```py
    model = MyModel()
    table = widgets.TableView()
    table.set_model(model)
    table.proxifier[::2, 2:].modify(xyz)
    table.show()
    # or
    indexer = (slice(None, None, 2), slice(2, None))
    proxy = itemmodels.SliceCheckableProxyModel(indexer=indexer)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "checkable"
    checkstate_changed = core.Signal(core.ModelIndex, bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._checked: set[tuple[int, int]] = set()

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return super().flags(index)
        if self.indexer_contains(index):
            return super().flags(index) | constants.IS_CHECKABLE
        return super().flags(index)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        key = self.get_index_key(index, include_column=True)
        if role == constants.CHECKSTATE_ROLE and self.indexer_contains(index):
            return key in self._checked
        return super().data(index, role)

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        key = self.get_index_key(index, include_column=True)
        if role == constants.CHECKSTATE_ROLE and self.indexer_contains(index):
            if is_checked := key in self._checked:
                self._checked.remove(key)
            else:
                self._checked.add(key)
            self.update_row(index.row())
            self.checkstate_changed.emit(index, not is_checked)
            return True

        return super().setData(index, role)


class SliceCheckableTreeProxyModel(itemmodels.SliceIdentityProxyModel):
    """Proxy model to make a tree model checkable.

    In contrast to SliceCheckableProxyModel, any checkstate change is propagated to
    parent and child indexes. (child indexes will also get the new checkstate, parent
    indexes will become partially checked if needed.)

    ### Example

    ```py
    model = MyModel()
    table = widgets.TreeView()
    table.set_model(model)
    table.proxifier[:, 0].set_checkable(tree=True)
    table.show()
    # or
    proxy = itemmodels.SliceCheckableTreeProxyModel(indexer=0)
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "checkable_tree"
    checkstate_changed = core.Signal(core.ModelIndex, constants.CheckState)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._checked: dict[tuple[int, int], constants.CheckState] = {}

    def flags(self, index: core.ModelIndex) -> constants.ItemFlag:
        if not index.isValid():
            return super().flags(index)
        if self.indexer_contains(index):
            # TODO: do i need to mess with ItemDataFlag.ItemIsTristate?
            return super().flags(index) | constants.IS_CHECKABLE
        return super().flags(index)

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if role == constants.CHECKSTATE_ROLE and self.indexer_contains(index):
            key = self.get_index_key(index, include_column=True)
            return self._checked.get(key, constants.CheckState.Unchecked)
        return super().data(index, role)

    def setData(
        self,
        index: core.ModelIndex,
        value: Any,
        role: constants.ItemDataRole = constants.EDIT_ROLE,
    ) -> bool:
        if role == constants.CHECKSTATE_ROLE and self.indexer_contains(index):
            self.set_checkstate(index, recursive=True)
        return super().setData(index, role)

    def set_checkstate(self, index: core.ModelIndex, recursive=False):
        key = self.get_index_key(index, include_column=True)
        current = self._checked.get(key)
        match current:
            case constants.CheckState.Checked:
                val = constants.CheckState.Unchecked
            case _:
                val = constants.CheckState.Checked
        self._checked[key] = val
        if recursive:
            self.set_child_states(index, val)
            self.set_parent_states(index)
        self.dataChanged.emit(index, index)
        self.checkstate_changed.emit(index, val)
        return True

    def set_parent_states(self, index: core.ModelIndex):
        indexes = []
        while (index := index.parent()).isValid():
            indexes.append(index)
        for idx in reversed(indexes):
            iterator = self.iter_tree(index)
            # next(iterator) # first one is ourself, throw it away
            states = []
            for index in iterator:
                state = index.data(constants.CHECKSTATE_ROLE)
                states.append(state)
            if all(state is True for state in states):
                val = True
            elif all(state is False for state in states):
                val = False
            else:
                val = constants.CheckState.PartiallyChecked
            key = self.get_index_key(idx, include_column=True)
            logger.debug(f"Setting {key} to {val}")
            self._checked[key] = val
            self.dataChanged.emit(idx, idx)
            self.checkstate_changed.emit(idx, val)

    def set_child_states(self, index, state):
        iterator = self.iter_tree(index)
        next(iterator)  # first one is ourself, throw it away
        for child_index in iterator:
            key = self.get_index_key(child_index, include_column=True)
            self._checked[key] = state
            self.dataChanged.emit(child_index, child_index)
            self.checkstate_changed.emit(child_index, state)


if __name__ == "__main__":
    from prettyqt import debugging, widgets

    app = widgets.app()
    tree = debugging.example_table()
    tree.proxifier[0, :].set_checkable()
    tree.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import functools

from prettyqt import widgets
from prettyqt.utils import get_repr


class GraphicsLayoutItemMixin:
    def __repr__(self):
        return get_repr(self)

    def __bool__(self):
        return not self.isEmpty()

    @functools.singledispatchmethod
    def set_size_policy(
        self,
        horizontal: widgets.sizepolicy.SizePolicyStr,
        vertical: widgets.sizepolicy.SizePolicyStr,
        control_type: widgets.sizepolicy.ControlTypeStr = "default",
    ):
        """Set the size policy.

        Args:
            horizontal: horizontal size policy
            vertical: vertical size policy
            control_type: control type associated with the widget
        """
        h_policy = widgets.sizepolicy.SIZE_POLICY[horizontal]
        v_policy = widgets.sizepolicy.SIZE_POLICY[vertical]
        c = widgets.sizepolicy.CONTROL_TYPE[control_type]
        self.setSizePolicy(h_policy, v_policy, c)

    @set_size_policy.register
    def _(self, policy: widgets.QSizePolicy):
        self.setSizePolicy(policy)

    def get_size_policy(self) -> widgets.SizePolicy:
        qpol = self.sizePolicy()
        return widgets.SizePolicy.clone(qpol)


class GraphicsLayoutItem(GraphicsLayoutItemMixin, widgets.QGraphicsLayoutItem):
    """Can be inherited to allow your custom items to be managed by layouts."""


if __name__ == "__main__":
    item = GraphicsLayoutItem()
    item.set_size_policy(widgets.SizePolicy())



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
import contextlib

from difflib import SequenceMatcher
import inspect
import logging
import pprint

from prettyqt import constants, core, gui, itemmodels


logger = logging.getLogger(__name__)


_ALL_PREDICATES = (
    inspect.isgenerator,
    inspect.istraceback,
    inspect.isframe,
    inspect.iscode,
    inspect.isabstract,
    inspect.isgetsetdescriptor,
    inspect.ismemberdescriptor,
)


TYPE_CHECK: dict[Callable, str] = {
    inspect.isasyncgen: "Async generator",
    inspect.ismethoddescriptor: "Method descriptor",
    inspect.ismethod: "Method",
    inspect.isfunction: "Function",
    inspect.isgeneratorfunction: "Generator function",
    inspect.isclass: "Class",
    inspect.ismodule: "Module",
    inspect.isdatadescriptor: "Data descriptor",
    inspect.isbuiltin: "Builtin",
}


class NameColumn(itemmodels.ColumnItem):
    name = "Name"
    doc = "The name of the object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj_name or "<root>"
            case constants.FOREGROUND_ROLE:
                return gui.QColor("green") if callable(item.obj) else None
            case constants.FONT_ROLE:
                if (
                    item.is_attribute
                    and item.obj_name.startswith("__")
                    and item.obj_name.endswith("__")
                ):
                    font = gui.QFont()
                    font.setItalic(True)
                    return font
            case constants.USER_ROLE:
                return item


class DescriptionColumn(itemmodels.ColumnItem):
    name = "Description"
    doc = "Description of the object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return next(
                    (v for k, v in TYPE_CHECK.items() if k(item.obj)), "Attribute"
                )


class PathColumn(itemmodels.ColumnItem):
    name = "Path"
    doc = "A path to the data: e.g. var[1]['a'].item"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj_path or "<root>"


class StrColumn(itemmodels.ColumnItem):
    name = "Str"
    doc = "The string representation of the object using the str() function."
    # line_wrap = "boundary_or_anywhere"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return str(item.obj)


class ReprColumn(itemmodels.ColumnItem):
    name = "Repr"
    doc = "The string representation of the object using the repr() function."
    # line_wrap = "boundary_or_anywhere"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return repr(item.obj)


class TypeColumn(itemmodels.ColumnItem):
    name = "Type"
    doc = "Type of the object determined using the builtin type() function."
    # line_wrap = "boundary_or_anywhere"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return str(type(item.obj))


class ClassColumn(itemmodels.ColumnItem):
    name = "Class"
    doc = "Class name of the object."
    # line_wrap = "boundary_or_anywhere"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return type(item.obj).__name__


class LengthColumn(itemmodels.ColumnItem):
    name = "Length"
    doc = "Length of the object if available."
    # line_wrap = "boundary_or_anywhere"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return str(len(item))


class IdColumn(itemmodels.ColumnItem):
    name = "Id"
    doc = "The identifier of the object calculated using the id() function."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return f"0x{id(item.obj):X}"


class AttributeColumn(itemmodels.ColumnItem):
    name = "Attribute"
    doc = "Whether the object is an attribute."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(bool(item.is_attribute))


class IsCallableColumn(itemmodels.ColumnItem):
    name = "Is callable"
    doc = "Whether the object is an callable (like functions or classes)."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(callable(item))


class IsRoutineColumn(itemmodels.ColumnItem):
    name = "Is routine"
    doc = "True if the object is a user-defined or built-in function or method."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(inspect.isroutine(item))


class IsBuiltinColumn(itemmodels.ColumnItem):
    name = "Builtin"
    doc = "True if the object is a user-defined or built-in function or method."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.CHECKSTATE_ROLE:
                return self.to_checkstate(inspect.isbuiltin(item))


class PredicateColumn(itemmodels.ColumnItem):
    name = "Predicate"

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return ", ".join(
                    pred.__name__ for pred in _ALL_PREDICATES if pred(item.obj)
                )


class PrettyPrintColumn(itemmodels.ColumnItem):
    name = "Pretty print"
    doc = "Pretty printed representation of the object using the pprint module."
    indent = 4

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                prettyprinter = pprint.PrettyPrinter(indent=self.indent)
                return prettyprinter.pformat(item.obj)


class DocStringColumn(itemmodels.ColumnItem):
    name = "Docstring"
    doc = "Docstring from source code."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return inspect.getdoc(item)


class CommentsColumn(itemmodels.ColumnItem):
    name = "Comments"
    doc = "Comments above the object's definition."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                return inspect.getcomments(item)


class ModuleColumn(itemmodels.ColumnItem):
    name = "Module"
    doc = "Module of given object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return inspect.getmodule(item)


class FileColumn(itemmodels.ColumnItem):
    name = "File"
    doc = "File of the given object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return inspect.getfile(item)


class SourceFileColumn(itemmodels.ColumnItem):
    name = "SourceFile"
    doc = "Source file of given object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return inspect.getsourcefile(item)


class SourceCodeColumn(itemmodels.ColumnItem):
    name = "SourceCode"
    doc = "Source code of given object."

    def get_data(self, item, role: constants.ItemDataRole = constants.DISPLAY_ROLE):
        match role:
            case constants.DISPLAY_ROLE:
                with contextlib.suppress(Exception):
                    return inspect.getsource(item)


class PythonObjectTreeItem(itemmodels.ColumnItemModel.TreeItem):
    """Tree node class that can be used to build trees of objects."""

    # __slots__ = ("obj_name", "obj_path", "is_attribute")

    def __init__(
        self,
        obj,
        name: str = "",
        obj_path="",
        is_attribute: bool = False,
        parent: PythonObjectTreeItem | None = None,
    ):
        super().__init__(obj, parent=parent)
        self.obj_name = name
        self.obj_path = str(obj_path)
        self.is_attribute = is_attribute


class PythonObjectTreeModel(itemmodels.ColumnItemModel):
    TreeItem = PythonObjectTreeItem
    IS_RECURSIVE = True
    ICON = "mdi.python"
    SUPPORTS = object
    COLUMNS = [
        NameColumn,
        DescriptionColumn,
        PathColumn,
        StrColumn,
        ReprColumn,
        LengthColumn,
        TypeColumn,
        ClassColumn,
        IdColumn,
        AttributeColumn,
        IsCallableColumn,
        IsRoutineColumn,
        IsBuiltinColumn,
        PredicateColumn,
        ModuleColumn,
        # FileColumn,
        # SourceFileColumn,
    ]

    def __init__(self, obj, parent=None):
        super().__init__(obj, self.COLUMNS, show_root=False, parent=parent)

    @classmethod
    def supports(cls, instance) -> bool:
        return True

    def get_path_for_index(self, index: core.ModelIndex) -> str:
        """Get the path for the object referenced by index.

        ### Example:
        ```
        An.example = {"a": [b, c, {"d": e}]} -> path of e: An.example["a"][2]["d"]
        ```
        """
        # TODO: not used yet, better rework ColumnItemModel first
        treeitem = index.data(constants.USER_ROLE)
        if treeitem is None:
            return
        prev_data = treeitem.obj
        pieces = []
        while (index := index.parent()).isValid():
            treeitem = index.data(constants.USER_ROLE)
            data = treeitem.obj
            match data:
                case Mapping():
                    for k, v in data.items():
                        if v is prev_data:
                            pieces.append(f"[{k!r}]")
                            break
                case Iterable():
                    pieces.append(f"[{data.index(prev_data)}]")
                case _:
                    # or should this be treeitem.obj_name?
                    pieces.append(f".{prev_data.__name__}")
            prev_data = data
        pieces.append(treeitem.obj_name)
        logger.info(pieces)
        return "".join(reversed(pieces))

    def _fetch_object_children(
        self, treeitem: PythonObjectTreeItem
    ) -> list[PythonObjectTreeItem]:
        """Fetch the children of a Python object.

        Returns: list of PythonObjectTreeItems
        """
        obj_children = []
        path_strings = []
        obj = treeitem.obj
        obj_path = treeitem.obj_path
        if isinstance(obj, list | tuple | set | frozenset):
            obj_children = [(str(i), j) for i, j in sorted(enumerate(obj))]
            path_strings = [
                f"{obj_path}[{i[0]}]" if obj_path else i[0] for i in obj_children
            ]
        elif isinstance(obj, Mapping):
            obj_children = list(obj.items())
            path_strings = [
                f"{obj_path}[{item[0]!r}]" if obj_path else item[0]
                for item in obj_children
            ]

        is_attr_list = [False] * len(obj_children)

        # Object attributes
        for attr_name, attr_value in sorted(inspect.getmembers(obj)):
            obj_children.append((attr_name, attr_value))
            path_strings.append(f"{obj_path}.{attr_name}" if obj_path else attr_name)
            is_attr_list.append(True)

        return [
            PythonObjectTreeItem(obj=val, name=name, obj_path=p, is_attribute=is_attr)
            for (name, val), p, is_attr in zip(obj_children, path_strings, is_attr_list)
        ]

    def _aux_refresh_tree(self, tree_index: core.ModelIndex):
        """Refresh the tree nodes recursively, auxiliary.

        If the underlying Python object has been changed, we don't want to delete the old
        tree model and create a new one from scratch because this loses all information
        about which nodes are fetched and expanded. Instead the old tree model is updated.
        Using the difflib from the standard library it is determined for a parent node
        which child nodes should be added or removed. This is done based on the node names
        only, not on the node contents (the underlying Python objects). Testing the
        underlying nodes for equality is potentially slow. It is faster to let the
        refreshNode function emit the dataChanged signal for all cells.
        """
        tree_item = self.data_by_index(tree_index)
        if not tree_item.children_fetched:
            return None
        old_items = tree_item.children
        new_items = self._fetch_object_children(tree_item)

        old_item_names = [(item.obj_name, item.is_attribute) for item in old_items]
        new_item_names = [(item.obj_name, item.is_attribute) for item in new_items]
        seq_matcher = SequenceMatcher(
            isjunk=None, a=old_item_names, b=new_item_names, autojunk=False
        )
        opcodes = seq_matcher.get_opcodes()

        logger.debug("(reversed) opcodes: %s", list(reversed(opcodes)))

        for tag, i1, i2, j1, j2 in reversed(opcodes):
            match tag:
                case "equal":
                    # when node names are equal is aux_refresh_tree called recursively.
                    for old_row, new_row in zip(range(i1, i2), range(j1, j2)):
                        old_items[old_row].obj = new_items[new_row].obj
                        child_index = self.index(old_row, 0, parent=tree_index)
                        self._aux_refresh_tree(child_index)

                case "replace":
                    # Remove the old item and insert the new. The old item may have child
                    # nodes which indices must be removed by Qt, otherwise it crashes.
                    first = i1  # row number of first that will be removed
                    last = i1 + i2 - 1  # row number of last element after insertion
                    with self.remove_rows(first, last, tree_index):
                        del tree_item.children[i1:i2]

                    first = i1  # row number of first element after insertion
                    last = i1 + j2 - j1 - 1  # row number of last element after insertion
                    with self.insert_rows(first, last, tree_index):
                        tree_item.insert_children(i1, new_items[j1:j2])

                case "delete":
                    first = i1  # row number of first that will be removed
                    last = i1 + i2 - 1  # row number of last element after insertion
                    with self.remove_rows(first, last, tree_index):
                        del tree_item.children[i1:i2]

                case "insert":
                    first = i1
                    last = i1 + j2 - j1 - 1
                    with self.insert_rows(first, last, tree_index):
                        tree_item.insert_children(i1, new_items[j1:j2])
                case _:
                    raise ValueError(f"Invalid tag: {tag}")

    def refresh_tree(self):
        if self._show_root:
            index = self.createIndex(0, 0, self.inspected_item)
        else:
            index = self.root_index()
        """Refresh the tree model from the underlying root object."""
        self._aux_refresh_tree(index)
        # Emit the dataChanged signal for all cells. This is faster than checking which
        # nodes have changed, which may be slow for some underlying Python objects.
        self.update_all()


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    with app.debug_mode():
        obj = dict(a=[0, 1, 2, 3, dict(b=5)], b={"test": {"test2": 5}})
        model = PythonObjectTreeModel(obj)
        obj_tree = widgets.TreeView()
        obj_tree.setRootIsDecorated(True)
        obj_tree.setAlternatingRowColors(True)
        obj_tree.set_model(model)
        obj_tree.selectionModel().currentRowChanged.connect(model.get_path_for_index)
        obj_tree.set_selection_behavior("rows")
        obj_tree.setUniformRowHeights(True)
        obj_tree.show()
        app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, gui
from prettyqt.utils import bidict, get_repr


FORMAT_TYPE = bidict(
    invalid=gui.QTextFormat.FormatType.InvalidFormat,
    block=gui.QTextFormat.FormatType.BlockFormat,
    char=gui.QTextFormat.FormatType.CharFormat,
    list=gui.QTextFormat.FormatType.ListFormat,
    frame=gui.QTextFormat.FormatType.FrameFormat,
    user=gui.QTextFormat.FormatType.UserFormat,
)

FormatTypeStr = Literal["invalid", "block", "char", "list", "frame", "user"]

OBJECT_TYPE = bidict(
    none=gui.QTextFormat.ObjectTypes.NoObject,
    image=gui.QTextFormat.ObjectTypes.ImageObject,
    table=gui.QTextFormat.ObjectTypes.TableObject,
    table_cell=gui.QTextFormat.ObjectTypes.TableCellObject,
    user=gui.QTextFormat.ObjectTypes.UserObject,
)

ObjectTypeStr = Literal["none", "image", "table", "table_cell", "user"]

PAGE_BREAK_FLAG = bidict(
    auto=gui.QTextFormat.PageBreakFlag.PageBreak_Auto,
    always_before=gui.QTextFormat.PageBreakFlag.PageBreak_AlwaysBefore,
    always_after=gui.QTextFormat.PageBreakFlag.PageBreak_AlwaysAfter,
)

PageBreakFlagStr = Literal["auto", "always_before", "always_after"]


class TextFormatMixin:
    def __getitem__(self, key: int):
        return self.property(key)

    def __setitem__(self, key: int, value):
        self.setProperty(key, value)

    def __contains__(self, key: int):
        return self.hasProperty(key)

    def __bool__(self):
        return self.isValid()

    def __repr__(self):
        return get_repr(self, self.type())

    def get_background(self) -> gui.Brush:
        return gui.Brush(self.background())

    def get_foreground(self) -> gui.Brush:
        return gui.Brush(self.foreground())

    def get_brush_property(self, property_id: int) -> gui.Brush:
        return gui.Brush(self.brushProperty(property_id))

    def get_color_property(self, property_id: int) -> gui.Color:
        return gui.Color(self.colorProperty(property_id))

    def get_pen_property(self, property_id: int) -> gui.Pen:
        return gui.Pen(self.penProperty(property_id))

    def set_layout_direction(
        self, direction: constants.LayoutDirectionStr | constants.LayoutDirection
    ):
        """Set layout direction.

        Args:
            direction: layout direction
        """
        self.setLayoutDirection(constants.LAYOUT_DIRECTION.get_enum_value(direction))

    def get_layout_direction(self) -> constants.LayoutDirectionStr:
        """Get the current layout direction.

        Returns:
            layout direction
        """
        return constants.LAYOUT_DIRECTION.inverse[self.layoutDirection()]

    def select_full_width(self, value: bool = True):
        prop = gui.QTextFormat.Property.FullWidthSelection
        self.setProperty(prop, value)  # type: ignore


class TextFormat(TextFormatMixin, gui.QTextFormat):
    pass


if __name__ == "__main__":
    fmt = TextFormat()



from __future__ import annotations

from prettyqt import gui


class Bitmap(gui.PixmapMixin, gui.QBitmap):
    """Monochrome (1-bit depth) pixmaps."""


if __name__ == "__main__":
    app = gui.app()
    image = Bitmap()
    bytes(image)
